static const char *RcsId = "";
//+=============================================================================
//
// file :         AH401B.cpp
//
// description :  C++ source for the AH401B and its commands. 
//                The class is derived from Device. It represents the
//                CORBA servant object which will be accessed from the
//                network. All commands which can be executed on the
//                AH401B are implemented in this file.
//
// project :      TANGO Device Server
//
//
// copyleft :   Sincrotrone Trieste S.C.p.A.
//              Strada Statale 14 - km 163,5 in AREA Science Park
//              34012 Basovizza (TS)
//              ITALY
//
//-=============================================================================
//
//  		This file is generated by POGO
//	(Program Obviously used to Generate tango Object)
//
//         (c) - Software Engineering Group - ESRF 
//=============================================================================



//===================================================================
//
//	The following table gives the correspondence
//	between commands and method name.
//
//  Command name        |  Method name
//	----------------------------------------
//  State               |  dev_state()
//  Status              |  dev_status()
//  SendCommand         |  send_command()
//  GetSnapshot         |  get_snapshot()
//  GetSamples          |  get_samples()
//  SetIntegrationTime  |  set_integration_time()
//  SetFullScaleRange   |  set_full_scale_range()
//  SetOneshotIntTime   |  set_oneshot_int_time()
//  GetCH1Data          |  get_ch1_data()
//  GetCH2Data          |  get_ch2_data()
//  GetCH3Data          |  get_ch3_data()
//  GetCH4Data          |  get_ch4_data()
//  AcquireBackground   |  acquire_background()
//
//===================================================================

//TODO: comandi inviati solo se NON in RUNNING

#include <tango.h>
#include <AH401B.h>
#include <AH401BClass.h>
#include <fcntl.h>
#include <sys/ioctl.h>
#include <sys/mman.h>
#include <check_get_rt_buffer.h>

#define _PUSH_EVENTS	0


// #define DEBUG

/*#ifndef DEBUG
  #ifndef dbgprintf
  static void dbgprintf(...) {}
  #endif
#else
#define dbgprintf printf
#endif
*/
namespace AH401B_ns
{

//+----------------------------------------------------------------------------
//
// method : 		AH401B::AH401B(string &s)
// 
// description : 	constructor for simulated AH401B
//
// in : - cl : Pointer to the DeviceClass object
//      - s : Device name 
//
//-----------------------------------------------------------------------------
AH401B::AH401B(Tango::DeviceClass *cl,string &s)
:TANGO_BASE_CLASS(cl,s.c_str())
{
	init_device();
}

AH401B::AH401B(Tango::DeviceClass *cl,const char *s)
:TANGO_BASE_CLASS(cl,s)
{
	init_device();
}

AH401B::AH401B(Tango::DeviceClass *cl,const char *s,const char *d)
:TANGO_BASE_CLASS(cl,s,d)
{
	init_device();
}
//+----------------------------------------------------------------------------
//
// method : 		AH401B::delete_device()
// 
// description : 	will be called at device destruction or at init command.
//
//-----------------------------------------------------------------------------
void AH401B::delete_device()
{
	//	Delete device's allocated object
	/* to avoid eth crash on shutdown */
	mutex->lock();
	sendCmd("ACQ OFF");
	getResponse();
	usleep(200000);
	sendCmd("TRG OFF");
	getResponse();
	mutex->unlock();

	delete attr_BunchNumber_read;
	delete attr_CH1Data_read;
	delete attr_CH2Data_read;
	delete attr_CH3Data_read;
	delete attr_CH4Data_read;
	delete attr_CH1_read;
	delete attr_CH2_read;
	delete attr_CH3_read;
	delete attr_CH4_read;
	delete attr_CH1Offset_read;
	delete attr_CH2Offset_read;
	delete attr_CH3Offset_read;
	delete attr_CH4Offset_read;
	delete attr_IntegrationTime_read;
	delete attr_Range_read;
	delete attr_NumAcq_read;

	if (picoammeter_deinit(picoammeter) < 0)
		INFO_STREAM << "AH401B::delete_device: Failed to remove realtime shared memory" << endl;

	delete_buffers();
	delete mutex;
	delete attr_HighVoltage_read;
	delete attr_HighVoltageSet_read;
	delete attr_Alarm1_read;
	delete attr_Alarm2_read;

}

//+----------------------------------------------------------------------------
//
// method : 		AH401B::init_device()
// 
// description : 	will be called at device initialization.
//
//-----------------------------------------------------------------------------
void AH401B::init_device()
{
	INFO_STREAM << "AH401B::AH401B() create device " << device_name << endl;
	flag_trg_upd = false;
	flag_itm_upd = false;
	flag_rng_upd = false;
	flag_naq_upd = false;
	flag_acq_upd = false;
	flag_hvs_upd = false;
	flag_hvs_set_upd = false;
	flag_alarm1_upd = false;
	flag_alarm2_upd = false;

	attr_Alarm1_write = 0; //TODO: temp
	attr_Alarm2_write = 0; //TODO: temp

	state_tout = STATE_TOUT;
	gettimeofday(&last_hv_set_time, NULL);
	gettimeofday(&last_hv_time,NULL);

	comm_error_mask = 0;
	// Initialise variables to default values
	//--------------------------------------------
	mutex = new omni_mutex::omni_mutex();
	get_device_property();
#if _PUSH_EVENTS
	set_change_event("State", true, false);		//attribute name, implemented, detect, if not setting detect=false have to set abs_change or rel_change for each attribute
#endif

	Tango::Database *db = new Tango::Database();
	try
	{
		Tango::DbData db_data;
		db_data.push_back((Tango::DbDatum("Host")));
		db_data.push_back((Tango::DbDatum("Port")));
		db->get_property("Database",db_data);
 
		db_data[0] >> host_rw;
		db_data[1] >> port_rw;
	}catch(Tango::DevFailed &e)
	{
		INFO_STREAM << __FUNCTION__ << " Error reading Database property='" << e.errors[0].desc << "'";
	}
	delete db; 

	get_configuration();	
	
	init_buffers();

	Tango::DevState st = Tango::UNKNOWN;
	set_state(st);

	picoammeter = 0;
	DEBUG_STREAM << "AH401B::AH401B: before picoammeter_init" << endl;
	if ((picoammeter = picoammeter_init()) == NULL) {
		WARN_STREAM << "AH401B::AH401B: Failed to initialize realtime shared memory, exit" << endl;
		exit(0);
	}

	unsigned char uc_cpu_ip[4];
	vector<unsigned int> parsed_ip_cpu;
	int i=0;
	string_explode(cpuIP, ".", &parsed_ip_cpu);
	if(parsed_ip_cpu.size() == 4)
	{
		for(vector<unsigned int>::iterator it=parsed_ip_cpu.begin(); it != parsed_ip_cpu.end(); it++)
			uc_cpu_ip[i++] = (unsigned char)*it;
	}
	vector<unsigned int> parsed_ip;
	i=0;
	string_explode(picoammeterIP, ".", &parsed_ip);
	if(parsed_ip.size() == 4)
	{
		for(vector<unsigned int>::iterator it=parsed_ip.begin(); it != parsed_ip.end(); it++)
			picoammeter_ip[i++] = (unsigned char)*it;
	}
	vector<unsigned int> parsed_mac;
	i=0;
	string_explode(picoammeterMAC, ":", &parsed_mac, true);
	if(parsed_mac.size() == 6)
	{
		for(vector<unsigned int>::iterator it=parsed_mac.begin(); it != parsed_mac.end(); it++)
		{
			picoammeter_mac[i++] = (unsigned char)*it;
		}
	}
	picoammeter_configure_eth(picoammeterIndex, picoammeter_mac, picoammeter_ip, uc_cpu_ip, picoammeterPort, cpuPort, picoammeter);

	for(int ch=0; ch<PICO_NUM_CHAN; ch++)
		picoammeter_configure_index(picoammeterIndex, rnmIndex, ch, picoammeter);

	attr_PAStatus_read = &statusArray[0];
	memset(statusArray, 0, STATUSARRAY_LENGTH);

	attr_BunchNumber_read = new Tango::DevLong();
	attr_CH1Data_read = new Tango::DevDouble();
	attr_CH2Data_read = new Tango::DevDouble();
	attr_CH3Data_read = new Tango::DevDouble();
	attr_CH4Data_read = new Tango::DevDouble();
	attr_CH1_read = new Tango::DevDouble();
	attr_CH2_read = new Tango::DevDouble();
	attr_CH3_read = new Tango::DevDouble();
	attr_CH4_read = new Tango::DevDouble();
	attr_CH1Offset_read = new Tango::DevDouble();
	attr_CH2Offset_read = new Tango::DevDouble();
	attr_CH3Offset_read = new Tango::DevDouble();
	attr_CH4Offset_read = new Tango::DevDouble();

	attr_IntegrationTime_read = new Tango::DevDouble();
	attr_Range_read = new Tango::DevLong();
	attr_NumAcq_read = new Tango::DevLong();
	attr_HighVoltage_read = new Tango::DevDouble();
	attr_HighVoltageSet_read = new Tango::DevDouble();
	attr_Alarm1_read = new Tango::DevLong();
	attr_Alarm2_read = new Tango::DevLong();

	attr_Range_write = range;
	statusArray[RNG] = range;	//TODO: needed to have the right status
	if(triggered)
		attr_NumAcq_write = numAcq;
	else
		attr_NumAcq_write = 0;
	attr_IntegrationTime_write = integrationTime;
	
	if (init_ah401b() == false)
		printf("Failed to initialize ah401b (picoammeter)\n");
	
	gettimeofday(&last_state_time, NULL);

	last_state_time.tv_sec -= 100;
	
	attr_Calibration_write = true;
	attr_Background_write = true;

}


//+----------------------------------------------------------------------------
//
// method : 		AH401B::get_device_property()
// 
// description : 	Read the device properties from database.
//
//-----------------------------------------------------------------------------
void AH401B::get_device_property()
{
	//	Initialize your default values here (if not done with  POGO).
	//------------------------------------------------------------------

	//	Read device properties from database.(Automatic code generation)
	//------------------------------------------------------------------
	Tango::DbData	dev_prop;
	dev_prop.push_back(Tango::DbDatum("BufferCH1Device"));
	dev_prop.push_back(Tango::DbDatum("BufferCH2Device"));
	dev_prop.push_back(Tango::DbDatum("BufferCH3Device"));
	dev_prop.push_back(Tango::DbDatum("BufferCH4Device"));
	dev_prop.push_back(Tango::DbDatum("CH1Cal"));
	dev_prop.push_back(Tango::DbDatum("CH2Cal"));
	dev_prop.push_back(Tango::DbDatum("CH3Cal"));
	dev_prop.push_back(Tango::DbDatum("CH4Cal"));
	dev_prop.push_back(Tango::DbDatum("IntegrationTime"));
	dev_prop.push_back(Tango::DbDatum("NSamples"));
	dev_prop.push_back(Tango::DbDatum("NumAcq"));
	dev_prop.push_back(Tango::DbDatum("Range"));
	dev_prop.push_back(Tango::DbDatum("BackgroundLevel"));
	dev_prop.push_back(Tango::DbDatum("PicoammeterIndex"));
	dev_prop.push_back(Tango::DbDatum("RnmIndex"));
	dev_prop.push_back(Tango::DbDatum("PicoammeterIP"));
	dev_prop.push_back(Tango::DbDatum("PicoammeterMAC"));
	dev_prop.push_back(Tango::DbDatum("PicoammeterPort"));
	dev_prop.push_back(Tango::DbDatum("CpuIP"));
	dev_prop.push_back(Tango::DbDatum("CpuPort"));
	dev_prop.push_back(Tango::DbDatum("Triggered"));
	dev_prop.push_back(Tango::DbDatum("HalfMode"));
	dev_prop.push_back(Tango::DbDatum("XpiMode"));

	//	Call database and extract values
	//--------------------------------------------
	if (Tango::Util::instance()->_UseDb==true)
		get_db_device()->get_property(dev_prop);
	Tango::DbDatum	def_prop, cl_prop;
	AH401BClass	*ds_class =
		(static_cast<AH401BClass *>(get_device_class()));
	int	i = -1;

	//	Try to initialize BufferCH1Device from class property
	cl_prop = ds_class->get_class_property(dev_prop[++i].name);
	if (cl_prop.is_empty()==false)	cl_prop  >>  bufferCH1Device;
	else {
		//	Try to initialize BufferCH1Device from default device value
		def_prop = ds_class->get_default_device_property(dev_prop[i].name);
		if (def_prop.is_empty()==false)	def_prop  >>  bufferCH1Device;
	}
	//	And try to extract BufferCH1Device value from database
	if (dev_prop[i].is_empty()==false)	dev_prop[i]  >>  bufferCH1Device;

	//	Try to initialize BufferCH2Device from class property
	cl_prop = ds_class->get_class_property(dev_prop[++i].name);
	if (cl_prop.is_empty()==false)	cl_prop  >>  bufferCH2Device;
	else {
		//	Try to initialize BufferCH2Device from default device value
		def_prop = ds_class->get_default_device_property(dev_prop[i].name);
		if (def_prop.is_empty()==false)	def_prop  >>  bufferCH2Device;
	}
	//	And try to extract BufferCH2Device value from database
	if (dev_prop[i].is_empty()==false)	dev_prop[i]  >>  bufferCH2Device;

	//	Try to initialize BufferCH3Device from class property
	cl_prop = ds_class->get_class_property(dev_prop[++i].name);
	if (cl_prop.is_empty()==false)	cl_prop  >>  bufferCH3Device;
	else {
		//	Try to initialize BufferCH3Device from default device value
		def_prop = ds_class->get_default_device_property(dev_prop[i].name);
		if (def_prop.is_empty()==false)	def_prop  >>  bufferCH3Device;
	}
	//	And try to extract BufferCH3Device value from database
	if (dev_prop[i].is_empty()==false)	dev_prop[i]  >>  bufferCH3Device;

	//	Try to initialize BufferCH4Device from class property
	cl_prop = ds_class->get_class_property(dev_prop[++i].name);
	if (cl_prop.is_empty()==false)	cl_prop  >>  bufferCH4Device;
	else {
		//	Try to initialize BufferCH4Device from default device value
		def_prop = ds_class->get_default_device_property(dev_prop[i].name);
		if (def_prop.is_empty()==false)	def_prop  >>  bufferCH4Device;
	}
	//	And try to extract BufferCH4Device value from database
	if (dev_prop[i].is_empty()==false)	dev_prop[i]  >>  bufferCH4Device;

	//	Try to initialize CH1Cal from class property
	cl_prop = ds_class->get_class_property(dev_prop[++i].name);
	if (cl_prop.is_empty()==false)	cl_prop  >>  cH1Cal;
	else {
		//	Try to initialize CH1Cal from default device value
		def_prop = ds_class->get_default_device_property(dev_prop[i].name);
		if (def_prop.is_empty()==false)	def_prop  >>  cH1Cal;
	}
	//	And try to extract CH1Cal value from database
	if (dev_prop[i].is_empty()==false)	dev_prop[i]  >>  cH1Cal;

	//	Try to initialize CH2Cal from class property
	cl_prop = ds_class->get_class_property(dev_prop[++i].name);
	if (cl_prop.is_empty()==false)	cl_prop  >>  cH2Cal;
	else {
		//	Try to initialize CH2Cal from default device value
		def_prop = ds_class->get_default_device_property(dev_prop[i].name);
		if (def_prop.is_empty()==false)	def_prop  >>  cH2Cal;
	}
	//	And try to extract CH2Cal value from database
	if (dev_prop[i].is_empty()==false)	dev_prop[i]  >>  cH2Cal;
	//	Try to initialize CH3Cal from class property
	cl_prop = ds_class->get_class_property(dev_prop[++i].name);
	if (cl_prop.is_empty()==false)	cl_prop  >>  cH3Cal;
	else {
		//	Try to initialize CH3Cal from default device value
		def_prop = ds_class->get_default_device_property(dev_prop[i].name);
		if (def_prop.is_empty()==false)	def_prop  >>  cH3Cal;
	}
	//	And try to extract CH3Cal value from database
	if (dev_prop[i].is_empty()==false)	dev_prop[i]  >>  cH3Cal;

	//	Try to initialize CH4Cal from class property
	cl_prop = ds_class->get_class_property(dev_prop[++i].name);
	if (cl_prop.is_empty()==false)	cl_prop  >>  cH4Cal;
	else {
		//	Try to initialize CH4Cal from default device value
		def_prop = ds_class->get_default_device_property(dev_prop[i].name);
		if (def_prop.is_empty()==false)	def_prop  >>  cH4Cal;
	}
	//	And try to extract CH4Cal value from database
	if (dev_prop[i].is_empty()==false)	dev_prop[i]  >>  cH4Cal;

	//	Try to initialize IntegrationTime from class property
	cl_prop = ds_class->get_class_property(dev_prop[++i].name);
	if (cl_prop.is_empty()==false)	cl_prop  >>  integrationTime;
	else {
		//	Try to initialize IntegrationTime from default device value
		def_prop = ds_class->get_default_device_property(dev_prop[i].name);
		if (def_prop.is_empty()==false)	def_prop  >>  integrationTime;
	}
	//	And try to extract IntegrationTime value from database
	if (dev_prop[i].is_empty()==false)	dev_prop[i]  >>  integrationTime;

	//	Try to initialize NSamples from class property
	cl_prop = ds_class->get_class_property(dev_prop[++i].name);
	if (cl_prop.is_empty()==false)	cl_prop  >>  nSamples;
	else {
		//	Try to initialize NSamples from default device value
		def_prop = ds_class->get_default_device_property(dev_prop[i].name);
		if (def_prop.is_empty()==false)	def_prop  >>  nSamples;
	}
	//	And try to extract NSamples value from database
	if (dev_prop[i].is_empty()==false)	dev_prop[i]  >>  nSamples;

	//	Try to initialize NumAcq from class property
	cl_prop = ds_class->get_class_property(dev_prop[++i].name);
	if (cl_prop.is_empty()==false)	cl_prop  >>  numAcq;
	else {
		//	Try to initialize NumAcq from default device value
		def_prop = ds_class->get_default_device_property(dev_prop[i].name);
		if (def_prop.is_empty()==false)	def_prop  >>  numAcq;
	}
	//	And try to extract NumAcq value from database
	if (dev_prop[i].is_empty()==false)	dev_prop[i]  >>  numAcq;

	//	Try to initialize Range from class property
	cl_prop = ds_class->get_class_property(dev_prop[++i].name);
	if (cl_prop.is_empty()==false)	cl_prop  >>  range;
	else {
		//	Try to initialize Range from default device value
		def_prop = ds_class->get_default_device_property(dev_prop[i].name);
		if (def_prop.is_empty()==false)	def_prop  >>  range;
	}
	//	And try to extract Range value from database
	if (dev_prop[i].is_empty()==false)	dev_prop[i]  >>  range;

	//	Try to initialize BackgroundLevel from class property
	cl_prop = ds_class->get_class_property(dev_prop[++i].name);
	if (cl_prop.is_empty()==false)	cl_prop  >>  backgroundLevel;
	else {
		//	Try to initialize BackgroundLevel from default device value
		def_prop = ds_class->get_default_device_property(dev_prop[i].name);
		if (def_prop.is_empty()==false)	def_prop  >>  backgroundLevel;
	}
	//	And try to extract BackgroundLevel value from database
	if (dev_prop[i].is_empty()==false)	dev_prop[i]  >>  backgroundLevel;

	//	Try to initialize PicoammeterIndex from class property
	cl_prop = ds_class->get_class_property(dev_prop[++i].name);
	if (cl_prop.is_empty()==false)	cl_prop  >>  picoammeterIndex;
	else {
		//	Try to initialize PicoammeterIndex from default device value
		def_prop = ds_class->get_default_device_property(dev_prop[i].name);
		if (def_prop.is_empty()==false)	def_prop  >>  picoammeterIndex;
	}
	//	And try to extract PicoammeterIndex value from database
	if (dev_prop[i].is_empty()==false)	dev_prop[i]  >>  picoammeterIndex;

	//	Try to initialize RnmIndex from class property
	cl_prop = ds_class->get_class_property(dev_prop[++i].name);
	if (cl_prop.is_empty()==false)	cl_prop  >>  rnmIndex;
	else {
		//	Try to initialize RnmIndex from default device value
		def_prop = ds_class->get_default_device_property(dev_prop[i].name);
		if (def_prop.is_empty()==false)	def_prop  >>  rnmIndex;
	}
	//	And try to extract RnmIndex value from database
	if (dev_prop[i].is_empty()==false)	dev_prop[i]  >>  rnmIndex;

	//	Try to initialize PicoammeterIP from class property
	cl_prop = ds_class->get_class_property(dev_prop[++i].name);
	if (cl_prop.is_empty()==false)	cl_prop  >>  picoammeterIP;
	else {
		//	Try to initialize PicoammeterIP from default device value
		def_prop = ds_class->get_default_device_property(dev_prop[i].name);
		if (def_prop.is_empty()==false)	def_prop  >>  picoammeterIP;
	}
	//	And try to extract PicoammeterIP value from database
	if (dev_prop[i].is_empty()==false)	dev_prop[i]  >>  picoammeterIP;

	//	Try to initialize PicoammeterMAC from class property
	cl_prop = ds_class->get_class_property(dev_prop[++i].name);
	if (cl_prop.is_empty()==false)	cl_prop  >>  picoammeterMAC;
	else {
		//	Try to initialize PicoammeterMAC from default device value
		def_prop = ds_class->get_default_device_property(dev_prop[i].name);
		if (def_prop.is_empty()==false)	def_prop  >>  picoammeterMAC;
	}
	//	And try to extract PicoammeterMAC value from database
	if (dev_prop[i].is_empty()==false)	dev_prop[i]  >>  picoammeterMAC;

	//	Try to initialize PicoammeterPort from class property
	cl_prop = ds_class->get_class_property(dev_prop[++i].name);
	if (cl_prop.is_empty()==false)	cl_prop  >>  picoammeterPort;
	else {
		//	Try to initialize PicoammeterPort from default device value
		def_prop = ds_class->get_default_device_property(dev_prop[i].name);
		if (def_prop.is_empty()==false)	def_prop  >>  picoammeterPort;
	}
	//	And try to extract PicoammeterPort value from database
	if (dev_prop[i].is_empty()==false)	dev_prop[i]  >>  picoammeterPort;

	//	Try to initialize CpuIP from class property
	cl_prop = ds_class->get_class_property(dev_prop[++i].name);
	if (cl_prop.is_empty()==false)	cl_prop  >>  cpuIP;
	else {
		//	Try to initialize CpuIP from default device value
		def_prop = ds_class->get_default_device_property(dev_prop[i].name);
		if (def_prop.is_empty()==false)	def_prop  >>  cpuIP;
	}
	//	And try to extract CpuIP value from database
	if (dev_prop[i].is_empty()==false)	dev_prop[i]  >>  cpuIP;

	//	Try to initialize CpuPort from class property
	cl_prop = ds_class->get_class_property(dev_prop[++i].name);
	if (cl_prop.is_empty()==false)	cl_prop  >>  cpuPort;
	else {
		//	Try to initialize CpuPort from default device value
		def_prop = ds_class->get_default_device_property(dev_prop[i].name);
		if (def_prop.is_empty()==false)	def_prop  >>  cpuPort;
	}
	//	And try to extract CpuPort value from database
	if (dev_prop[i].is_empty()==false)	dev_prop[i]  >>  cpuPort;

	//	Try to initialize Triggered from class property
	cl_prop = ds_class->get_class_property(dev_prop[++i].name);
	if (cl_prop.is_empty()==false)	cl_prop  >>  triggered;
	else {
		//	Try to initialize Triggered from default device value
		def_prop = ds_class->get_default_device_property(dev_prop[i].name);
		if (def_prop.is_empty()==false)	def_prop  >>  triggered;
	}
	//	And try to extract Triggered value from database
	if (dev_prop[i].is_empty()==false)	dev_prop[i]  >>  triggered;

	//	Try to initialize HalfMode from class property
	cl_prop = ds_class->get_class_property(dev_prop[++i].name);
	if (cl_prop.is_empty()==false)	cl_prop  >>  halfMode;
	else {
		//	Try to initialize HalfMode from default device value
		def_prop = ds_class->get_default_device_property(dev_prop[i].name);
		if (def_prop.is_empty()==false)	def_prop  >>  halfMode;
	}
	//	And try to extract HalfMode value from database
	if (dev_prop[i].is_empty()==false)	dev_prop[i]  >>  halfMode;

	//	Try to initialize XpiMode from class property
	cl_prop = ds_class->get_class_property(dev_prop[++i].name);
	if (cl_prop.is_empty()==false)	cl_prop  >>  xpiMode;
	else {
		//	Try to initialize XpiMode from default device value
		def_prop = ds_class->get_default_device_property(dev_prop[i].name);
		if (def_prop.is_empty()==false)	def_prop  >>  xpiMode;
	}
	//	And try to extract XpiMode value from database
	if (dev_prop[i].is_empty()==false)	dev_prop[i]  >>  xpiMode;


	//	End of Automatic code generation
	//------------------------------------------------------------------

}
//+----------------------------------------------------------------------------
//
// method : 		AH401B::always_executed_hook()
// 
// description : 	method always executed before any command is executed
//
//-----------------------------------------------------------------------------
void AH401B::always_executed_hook()
{
	int stat=-1;
	struct timeval curr_state_time;

	//DEBUG_STREAM << "AH401B::always_executed_hook entering comm_error_mask="<<(int)comm_error_mask;
	gettimeofday(&curr_state_time, NULL);
	double diff_time_s = fabs((curr_state_time.tv_sec - last_state_time.tv_sec) + ((double)(curr_state_time.tv_usec - last_state_time.tv_usec))/1000000.0);
/*	//moved inside read_HighVoltage
	double diff_time_hv_s = fabs((curr_state_time.tv_sec - last_hv_time.tv_sec) + ((double)(curr_state_time.tv_usec - last_hv_time.tv_usec))/1000000.0);
	double diff_time_hv_set_s = fabs((curr_state_time.tv_sec - last_hv_set_time.tv_sec) + ((double)(curr_state_time.tv_usec - last_hv_set_time.tv_usec))/1000000.0);*/

	if ((diff_time_s > state_tout) || 
		flag_trg_upd || flag_itm_upd || flag_rng_upd || flag_naq_upd || flag_acq_upd || flag_alarm1_upd || flag_alarm2_upd)
	{
		try {
			update_StatArray(flag_itm_upd, flag_rng_upd, flag_naq_upd, flag_alarm1_upd, flag_alarm2_upd);
		}
		catch(Tango::DevFailed &e)
		{
			if(get_state() != Tango::FAULT)
			{
				Tango::DevState st = Tango::FAULT;
				set_state(st);
#if _PUSH_EVENTS
				try {
					push_change_event("state", &st);
				} catch(Tango::DevFailed &e)
				{
					INFO_STREAM << "AH401B::always_executed_hook() error pushing state="<<e.errors[0].desc<<endl;
				}
#endif
			}
			set_status("Can't read device status");
		}
		flag_trg_upd = false;
		flag_itm_upd = false;
		flag_rng_upd = false;
		flag_naq_upd = false;
		flag_acq_upd = false;
		flag_alarm1_upd = false;
		flag_alarm2_upd = false;
	}

/*	//moved inside read_HighVoltage
	if (xpiMode && ((diff_time_hv_s > HV_TOUT) || (flag_hvs_upd && ((diff_time_hv_set_s > 1.7 && diff_time_hv_set_s < 2.1) || (diff_time_hv_set_s > 6.0 && diff_time_hv_set_s < 6.3)  || (diff_time_hv_set_s > 14.0 && diff_time_hv_set_s < 14.4) || (diff_time_hv_set_s > 23.0)))) || flag_hvs_set_upd)
	{
		try {
			update_HV(flag_hvs_upd, flag_hvs_set_upd);
		}
		catch(Tango::DevFailed &e)
		{

		}
		flag_hvs_set_upd = false;
		if((diff_time_hv_s > HV_TOUT) || (diff_time_hv_set_s > 23.0))
			flag_hvs_upd = false;
	}*/

	stat = picoammeter->pico[picoammeterIndex].stat_status;

	if (stat & PICO_STAT_NOT_UPDATING)
	{
		//DEBUG_STREAM << "PICO_STAT_NOT_UPDATING comm_error_mask="<<(int)comm_error_mask;
		if(get_state() == Tango::ON)
		{
			DEBUG_STREAM << "PICO_STAT_NOT_UPDATING Tango::ON";
			//if(get_state() != Tango::FAULT)
			{
				Tango::DevState st = Tango::FAULT;
				set_state(st);
#if _PUSH_EVENTS
				try {
					push_change_event("state", &st);
				} catch(Tango::DevFailed &e)
				{
					INFO_STREAM << "AH401B::always_executed_hook() error pushing state="<<e.errors[0].desc<<endl;
				}
#endif
			}
			set_status("Trigger missing or hw fault");
			//check if the reason for not updating is acquisition stopped
			update_StatArray(flag_itm_upd, flag_rng_upd, flag_naq_upd, flag_alarm1_upd, flag_alarm2_upd);
			state_tout = STATE_TOUT_2;
		}
		else if(get_state() == Tango::FAULT && comm_error_mask == (COMM_ERROR+COMM_OK))
		{
			DEBUG_STREAM << "PICO_STAT_NOT_UPDATING Tango::FAULT comm_error_mask="<<(int)comm_error_mask;
			comm_error_mask = 0;	//reset
			//just recovered from FAULT
			init_ah401b();
			try {
				update_StatArray(flag_itm_upd, flag_rng_upd, flag_naq_upd, flag_alarm1_upd, flag_alarm2_upd);
			}
			catch(Tango::DevFailed &e)
			{
				if(get_state() != Tango::FAULT)
				{
					Tango::DevState st = Tango::FAULT;
					set_state(st);
#if _PUSH_EVENTS
					try {
						push_change_event("state", &st);
					} catch(Tango::DevFailed &e)
					{
						INFO_STREAM << "AH401B::always_executed_hook() error pushing state="<<e.errors[0].desc<<endl;
					}
#endif
				}
				set_status("Can't read device status");
				return;
			}
		}
		else if(state_tout != STATE_TOUT_2)
		{
			DEBUG_STREAM << "PICO_STAT_NOT_UPDATING setting fast state timeout="<<(int)STATE_TOUT_2;
			state_tout = STATE_TOUT_2;
		}
	}
	else
	{
		if (get_state() != Tango::ON && get_state() != Tango::STANDBY && get_state() != Tango::OFF)
		{
			/*if(get_state() == Tango::FAULT)
			{
				//DEBUG_STREAM << "PICO_STAT_UPDATING Tango::FAULT";
				//just recovered from FAULT
				init_ah401b();
				try {
					update_StatArray(flag_itm_upd, flag_rng_upd, flag_naq_upd);
				}
				catch(Tango::DevFailed &e)
				{
					if(get_state() != Tango::FAULT)
					{
						Tango::DevState st = Tango::FAULT;
						set_state(st);
#if _PUSH_EVENTS
						try {
							push_change_event("state", &st);
						} catch(Tango::DevFailed &e)
						{
							INFO_STREAM << "AH401B::always_executed_hook() error pushing state="<<e.errors[0].desc<<endl;
						}
#endif
					}
					set_status("Can't read device status");
					return;
				}
			}*/
			if((statusArray[TRG] = 0 && triggered) || (statusArray[ACQ] = 0 && !triggered))
			{
				if(get_state() != Tango::STANDBY)
				{
					Tango::DevState st = Tango::STANDBY;
					set_state(st);
#if _PUSH_EVENTS
					try {
						push_change_event("state", &st);
					} catch(Tango::DevFailed &e)
					{
						INFO_STREAM << "AH401B::always_executed_hook() error pushing state="<<e.errors[0].desc<<endl;
					}
#endif
				}
				set_status("Not acquiring");
			}
			else if((statusArray[TRG] = 1 && triggered) || (statusArray[ACQ] = 1 && !triggered))
			{
				if(get_state() != Tango::ON)
				{
					Tango::DevState st = Tango::ON;
					set_state(st);
#if _PUSH_EVENTS
					try {
						push_change_event("state", &st);
					} catch(Tango::DevFailed &e)
					{
						INFO_STREAM << "AH401B::always_executed_hook() error pushing state="<<e.errors[0].desc<<endl;
					}
#endif
				}
				char tmp_stat[256];
				if(triggered)
				{
					sprintf(tmp_stat, "Acquiring, rng=%s", (statusArray[RNG] != -1) ? ranges[statusArray[RNG]] : "unknown");
				}
				else
				{
					double free_freq = 1000.0 / attr_IntegrationTime_write;
					free_freq = halfMode ? (free_freq/2) : free_freq;
					sprintf(tmp_stat, "Acquiring (not triggered) at %3.1f Hz, rng=%s", free_freq, (statusArray[RNG] != -1) ? ranges[statusArray[RNG]] : "unknown");
				}
				set_status(tmp_stat);
			}
		}
	}
}
//+----------------------------------------------------------------------------
//
// method : 		AH401B::read_attr_hardware
// 
// description : 	Hardware acquisition for attributes.
//
//-----------------------------------------------------------------------------
void AH401B::read_attr_hardware(vector<long> &attr_list)
{
	//DEBUG_STREAM << "AH401B::read_attr_hardware(vector<long> &attr_list) entering... "<< endl;
	//	Add your own code here
}
//+----------------------------------------------------------------------------
//
// method : 		AH401B::read_BunchNumber
// 
// description : 	Extract real attribute values for BunchNumber acquisition result.
//
//-----------------------------------------------------------------------------
void AH401B::read_BunchNumber(Tango::Attribute &attr)
{
	//DEBUG_STREAM << "AH401B::read_BunchNumber(Tango::Attribute &attr) entering... "<< endl;
	unsigned long bunchnumber;
	picoammeter_read_bunchnum(&bunchnumber, picoammeter);
	*attr_BunchNumber_read = bunchnumber;
	attr.set_value(attr_BunchNumber_read);		

}

//+----------------------------------------------------------------------------
//
// method : 		AH401B::read_Calibration
// 
// description : 	Extract real attribute values for Calibration acquisition result.
//
//-----------------------------------------------------------------------------
void AH401B::read_Calibration(Tango::Attribute &attr)
{
	//DEBUG_STREAM << "AH401B::read_Calibration(Tango::Attribute &attr) entering... "<< endl;

	attr.set_value(&attr_Calibration_write);		
	
}

//+----------------------------------------------------------------------------
//
// method : 		AH401B::write_Calibration
// 
// description : 	Write Calibration attribute values to hardware.
//
//-----------------------------------------------------------------------------
void AH401B::write_Calibration(Tango::WAttribute &attr)
{
	DEBUG_STREAM << "AH401B::write_Calibration(Tango::WAttribute &attr) entering... "<< endl;

	attr.get_write_value(attr_Calibration_write);	
	
}

//+----------------------------------------------------------------------------
//
// method : 		AH401B::read_Background
// 
// description : 	Extract real attribute values for Background acquisition result.
//
//-----------------------------------------------------------------------------
void AH401B::read_Background(Tango::Attribute &attr)
{
	//DEBUG_STREAM << "AH401B::read_Background(Tango::Attribute &attr) entering... "<< endl;

	attr.set_value(&attr_Background_write);		
	
}

//+----------------------------------------------------------------------------
//
// method : 		AH401B::write_Background
// 
// description : 	Write Background attribute values to hardware.
//
//-----------------------------------------------------------------------------
void AH401B::write_Background(Tango::WAttribute &attr)
{
	DEBUG_STREAM << "AH401B::write_Background(Tango::WAttribute &attr) entering... "<< endl;

	attr.get_write_value(attr_Background_write);

}


//+----------------------------------------------------------------------------
//
// method : 		AH401B::read_IntegrationTime
// 
// description : 	Extract real attribute values for IntegrationTime acquisition result.
//
//-----------------------------------------------------------------------------
void AH401B::read_IntegrationTime(Tango::Attribute &attr)
{
	//DEBUG_STREAM << "AH401B::read_IntegrationTime(Tango::Attribute &attr) entering... "<< endl;
	//*attr_IntegrationTime_read = statusArray[ITM];
	attr.set_value(attr_IntegrationTime_read);
	
}

//+----------------------------------------------------------------------------
//
// method : 		AH401B::write_IntegrationTime
// 
// description : 	Write IntegrationTime attribute values to hardware.
//
//-----------------------------------------------------------------------------
void AH401B::write_IntegrationTime(Tango::WAttribute &attr)
{
	DEBUG_STREAM << "AH401B::write_IntegrationTime(Tango::WAttribute &attr) entering... "<< endl;
	
	attr.get_write_value(attr_IntegrationTime_write);

	set_integration_time(attr_IntegrationTime_write);
	
}

//+----------------------------------------------------------------------------
//
// method : 		AH401B::read_Range
// 
// description : 	Extract real attribute values for Range acquisition result.
//
//-----------------------------------------------------------------------------
void AH401B::read_Range(Tango::Attribute &attr)
{
	//DEBUG_STREAM << "AH401B::read_Range(Tango::Attribute &attr) entering... "<< endl;
	*attr_Range_read = statusArray[RNG];
	attr.set_value(attr_Range_read);	
	
}

//+----------------------------------------------------------------------------
//
// method : 		AH401B::write_Range
// 
// description : 	Write Range attribute values to hardware.
//
//-----------------------------------------------------------------------------
void AH401B::write_Range(Tango::WAttribute &attr)
{
	DEBUG_STREAM << "AH401B::write_Range(Tango::WAttribute &attr) entering... "<< endl;
	
	attr.get_write_value(attr_Range_write);
	
	if (attr_Range_write < 0)
		attr_Range_write = 0;
		
	if (attr_Range_write > 7)
		attr_Range_write = 7;		
		
	set_full_scale_range((short)attr_Range_write);

}

//+----------------------------------------------------------------------------
//
// method : 		AH401B::read_HighVoltage
// 
// description : 	Extract real attribute values for HighVoltage acquisition result.
//
//-----------------------------------------------------------------------------
void AH401B::read_HighVoltage(Tango::Attribute &attr)
{
	//DEBUG_STREAM << "AH401B::read_HighVoltage(Tango::Attribute &attr) entering... "<< endl;
	if(!xpiMode)
		Tango::Except::throw_exception(
			(const char *) "Not Allowed",
			(const char *) "Allowed only for Xpi",
			(const char *)__FUNCTION__,Tango::ERR);
	struct timeval curr_state_time;
	gettimeofday(&curr_state_time, NULL);
	double diff_time_hv_s = fabs((curr_state_time.tv_sec - last_hv_time.tv_sec) + ((double)(curr_state_time.tv_usec - last_hv_time.tv_usec))/1000000.0);
	double diff_time_hv_set_s = fabs((curr_state_time.tv_sec - last_hv_set_time.tv_sec) + ((double)(curr_state_time.tv_usec - last_hv_set_time.tv_usec))/1000000.0);

	if (xpiMode && ((diff_time_hv_s > HV_TOUT) || (flag_hvs_upd && ((diff_time_hv_set_s > 1.7 && diff_time_hv_set_s < 2.1) || (diff_time_hv_set_s > 6.0 && diff_time_hv_set_s < 6.3)  || (diff_time_hv_set_s > 14.0 && diff_time_hv_set_s < 14.4) || (diff_time_hv_set_s > 23.0)))) || flag_hvs_set_upd)
	{
		try {
			update_HV(flag_hvs_upd, flag_hvs_set_upd);
		}
		catch(Tango::DevFailed &e)
		{

		}
		flag_hvs_set_upd = false;
		if((diff_time_hv_s > HV_TOUT) || (diff_time_hv_set_s > 23.0))
			flag_hvs_upd = false;
	}
	attr.set_value(attr_HighVoltage_read);
}

//+----------------------------------------------------------------------------
//
// method : 		AH401B::write_HighVoltaget
// 
// description : 	Write HighVoltage attribute values to hardware.
//
//-----------------------------------------------------------------------------
void AH401B::write_HighVoltage(Tango::WAttribute &attr)
{
	DEBUG_STREAM << "AH401B::write_HighVoltage(Tango::WAttribute &attr) entering... "<< endl;
	if(!xpiMode)
		Tango::Except::throw_exception(
			(const char *) "Not Allowed",
			(const char *) "Allowed only for Xpi",
			(const char *)__FUNCTION__,Tango::ERR);
	attr.get_write_value(attr_HighVoltage_write);
	set_high_volt(attr_HighVoltage_write);

}

//+----------------------------------------------------------------------------
//
// method : 		AH401B::read_HighVoltageSet
// 
// description : 	Extract real attribute values for HighVoltageSet acquisition result.
//
//-----------------------------------------------------------------------------
void AH401B::read_HighVoltageSet(Tango::Attribute &attr)
{
	//DEBUG_STREAM << "AH401B::read_HighVoltageSet(Tango::Attribute &attr) entering... "<< endl;
	if(!xpiMode)
		Tango::Except::throw_exception(
			(const char *) "Not Allowed",
			(const char *) "Allowed only for Xpi",
			(const char *)__FUNCTION__,Tango::ERR);
	
	struct timeval curr_state_time;
	gettimeofday(&curr_state_time, NULL);
	double diff_time_hv_s = fabs((curr_state_time.tv_sec - last_hv_time.tv_sec) + ((double)(curr_state_time.tv_usec - last_hv_time.tv_usec))/1000000.0);
	double diff_time_hv_set_s = fabs((curr_state_time.tv_sec - last_hv_set_time.tv_sec) + ((double)(curr_state_time.tv_usec - last_hv_set_time.tv_usec))/1000000.0);

	if (xpiMode && ((diff_time_hv_s > HV_TOUT) || (flag_hvs_upd && ((diff_time_hv_set_s > 1.7 && diff_time_hv_set_s < 2.1) || (diff_time_hv_set_s > 6.0 && diff_time_hv_set_s < 6.3)  || (diff_time_hv_set_s > 14.0 && diff_time_hv_set_s < 14.4) || (diff_time_hv_set_s > 23.0)))) || flag_hvs_set_upd)
	{
		try {
			update_HV(flag_hvs_upd, flag_hvs_set_upd);
		}
		catch(Tango::DevFailed &e)
		{

		}
		flag_hvs_set_upd = false;
		if((diff_time_hv_s > HV_TOUT) || (diff_time_hv_set_s > 23.0))
			flag_hvs_upd = false;
	}
	attr.set_value(attr_HighVoltageSet_read);
}

//+----------------------------------------------------------------------------
//
// method : 		AH401B::read_Alarm1
// 
// description : 	Extract real attribute values for Alarm1 acquisition result.
//
//-----------------------------------------------------------------------------
void AH401B::read_Alarm1(Tango::Attribute &attr)
{
	//DEBUG_STREAM << "AH401B::read_Alarm1(Tango::Attribute &attr) entering... "<< endl;
	if(!xpiMode)
		Tango::Except::throw_exception(
			(const char *) "Not Allowed",
			(const char *) "Allowed only for Xpi",
			(const char *)__FUNCTION__,Tango::ERR);

	attr.set_value(attr_Alarm1_read);
}

//+----------------------------------------------------------------------------
//
// method : 		AH401B::write_Alarm1
// 
// description : 	Write Alarm1 attribute values to hardware.
//
//-----------------------------------------------------------------------------
void AH401B::write_Alarm1(Tango::WAttribute &attr)
{
	DEBUG_STREAM << "AH401B::write_Alarm1(Tango::WAttribute &attr) entering... "<< endl;
	if(!xpiMode)
		Tango::Except::throw_exception(
			(const char *) "Not Allowed",
			(const char *) "Allowed only for Xpi",
			(const char *)__FUNCTION__,Tango::ERR);
	attr.get_write_value(attr_Alarm1_write);
	set_alarm1(attr_Alarm1_write);
}

//+----------------------------------------------------------------------------
//
// method : 		AH401B::read_Alarm2
// 
// description : 	Extract real attribute values for Alarm2 acquisition result.
//
//-----------------------------------------------------------------------------
void AH401B::read_Alarm2(Tango::Attribute &attr)
{
	//DEBUG_STREAM << "AH401B::read_Alarm2(Tango::Attribute &attr) entering... "<< endl;
	if(!xpiMode)
		Tango::Except::throw_exception(
			(const char *) "Not Allowed",
			(const char *) "Allowed only for Xpi",
			(const char *)__FUNCTION__,Tango::ERR);

	attr.set_value(attr_Alarm2_read);
}

//+----------------------------------------------------------------------------
//
// method : 		AH401B::write_Alarm2
// 
// description : 	Write Alarm2 attribute values to hardware.
//
//-----------------------------------------------------------------------------
void AH401B::write_Alarm2(Tango::WAttribute &attr)
{
	DEBUG_STREAM << "AH401B::write_Alarm2(Tango::WAttribute &attr) entering... "<< endl;
	if(!xpiMode)
		Tango::Except::throw_exception(
			(const char *) "Not Allowed",
			(const char *) "Allowed only for Xpi",
			(const char *)__FUNCTION__,Tango::ERR);
	attr.get_write_value(attr_Alarm2_write);
	set_alarm2(attr_Alarm2_write);
}

//+----------------------------------------------------------------------------
//
// method : 		AH401B::read_NumAcq
// 
// description : 	Extract real attribute values for NumAcq acquisition result.
//
//-----------------------------------------------------------------------------
void AH401B::read_NumAcq(Tango::Attribute &attr)
{
	//DEBUG_STREAM << "AH401B::read_NumAcq(Tango::Attribute &attr) entering... "<< endl;
	*attr_NumAcq_read = statusArray[NAQ];
	attr.set_value(attr_NumAcq_read);	
	
}

//+----------------------------------------------------------------------------
//
// method : 		AH401B::write_NumAcq
// 
// description : 	Write NumAcq attribute values to hardware.
//
//-----------------------------------------------------------------------------
void AH401B::write_NumAcq(Tango::WAttribute &attr)
{
	DEBUG_STREAM << "AH401B::write_NumAcq(Tango::WAttribute &attr) entering... "<< endl;

	attr.get_write_value(attr_NumAcq_write);	

	set_num_acq(attr_NumAcq_write);
	
}

//+----------------------------------------------------------------------------
//
// method : 		AH401B::read_CH1Data
// 
// description : 	Extract real attribute values for CH1Data acquisition result.
//
//-----------------------------------------------------------------------------
void AH401B::read_CH1Data(Tango::Attribute &attr)
{
	//DEBUG_STREAM << "AH401B::read_CH1Data(Tango::Attribute &attr) entering... "<< endl;
	
	long ret,data;
	unsigned long bunchnumber;
	struct timeval buf_time,cur_time;
	
	gettimeofday(&cur_time,NULL);
	
	ret = read_last(&data, &buf_time, &bunchnumber, buf[0]);
	double bt = buf_time.tv_sec + ((double)buf_time.tv_usec)/1000000;
	double ct = cur_time.tv_sec + ((double)cur_time.tv_usec)/1000000;
	//DEBUG_STREAM << "AH401B::read_CH1Data() buf= "<<hex<<(unsigned int)buf[0]<<dec<<" buf_time.tv_sec="<<(unsigned int)buf_time.tv_sec <<" data="<<(double)data<<endl;
	//if (((cur_time.tv_sec-buf_time.tv_sec)*1000000+(cur_time.tv_usec-buf_time.tv_usec)) > AH401B_TIMEOUT)
	if ((ct-bt)*1000000 > AH401B_TIMEOUT)
	{
		//DEBUG_STREAM << "AH401B::read_CH1Data() TIMEOUT buf time="<<fixed<<bt<<" cur time="<<ct<<" diff="<<ct-bt<<" data="<<(double)data<<endl;
		Tango::Except::throw_exception(
			(const char *) "Failed to retreive data",
			(const char *) "Timeout",
			(const char *) "AH401B::read_CH1Data()",Tango::ERR);
	}
	//DEBUG_STREAM << "AH401B::read_CH1Data() OK buf time="<<fixed<<bt<<" cur time="<<ct<<" diff="<<ct-bt<<" data="<<(double)data<<endl;
	*attr_CH1Data_read = (double) data;
	if (attr_Background_write)
		*attr_CH1Data_read -= background_level[0];
	if (attr_Calibration_write)
		*attr_CH1Data_read *= cal_matrix[0][attr_Range_write];
	
	attr.set_value(attr_CH1Data_read);	
	
}

//+----------------------------------------------------------------------------
//
// method : 		AH401B::read_CH2Data
// 
// description : 	Extract real attribute values for CH2Data acquisition result.
//
//-----------------------------------------------------------------------------
void AH401B::read_CH2Data(Tango::Attribute &attr)
{
	//DEBUG_STREAM << "AH401B::read_CH2Data(Tango::Attribute &attr) entering... "<< endl;
	
	long ret,data;
	unsigned long bunchnumber;
	struct timeval buf_time,cur_time;
	
	gettimeofday(&cur_time,NULL);
	
	ret = read_last(&data, &buf_time, &bunchnumber, buf[1]);
	double bt = buf_time.tv_sec + ((double)buf_time.tv_usec)/1000000;
	double ct = cur_time.tv_sec + ((double)cur_time.tv_usec)/1000000;	
	//DEBUG_STREAM << "AH401B::read_CH2Data() buf= "<<hex<<(unsigned int)buf[1]<<dec<<" buf_time.tv_sec="<<(unsigned int)buf_time.tv_sec <<" data="<<(double)data<<endl;
	//if (((cur_time.tv_sec-buf_time.tv_sec)*1000000+(cur_time.tv_usec-buf_time.tv_usec)) > AH401B_TIMEOUT)
	if ((ct-bt)*1000000 > AH401B_TIMEOUT)
	{
		//DEBUG_STREAM << "AH401B::read_CH2Data() TIMEOUT buf time="<<fixed<<bt<<" cur time="<<ct<<" diff="<<ct-bt<<" data="<<(double)data<<endl;
		Tango::Except::throw_exception(
			(const char *) "Failed to retreive data",
			(const char *) "Timeout",
			(const char *) "AH401B::read_CH2Data()",Tango::ERR);
	}
	//DEBUG_STREAM << "AH401B::read_CH2Data() OK buf time="<<fixed<<bt<<" cur time="<<ct<<" diff="<<ct-bt<<" data="<<(double)data<<endl;

	*attr_CH2Data_read = (double) data;
	if (attr_Background_write)
		*attr_CH2Data_read -= background_level[1];
	if (attr_Calibration_write)
		*attr_CH2Data_read *= cal_matrix[1][attr_Range_write];			
	
	attr.set_value(attr_CH2Data_read);		
}

//+----------------------------------------------------------------------------
//
// method : 		AH401B::read_CH3Data
// 
// description : 	Extract real attribute values for CH3Data acquisition result.
//
//-----------------------------------------------------------------------------
void AH401B::read_CH3Data(Tango::Attribute &attr)
{
	//DEBUG_STREAM << "AH401B::read_CH3Data(Tango::Attribute &attr) entering... "<< endl;

	long ret,data;
	unsigned long bunchnumber;
	struct timeval buf_time,cur_time;
	
	gettimeofday(&cur_time,NULL);

	ret = read_last(&data, &buf_time, &bunchnumber, buf[2]);
	double bt = buf_time.tv_sec + ((double)buf_time.tv_usec)/1000000;
	double ct = cur_time.tv_sec + ((double)cur_time.tv_usec)/1000000;	
	//DEBUG_STREAM << "AH401B::read_CH3Data() buf= "<<hex<<(unsigned int)buf[2]<<dec<<" buf_time.tv_sec="<<(unsigned int)buf_time.tv_sec <<" data="<<(double)data<<endl;
	//if (((cur_time.tv_sec-buf_time.tv_sec)*1000000+(cur_time.tv_usec-buf_time.tv_usec)) > AH401B_TIMEOUT)
	if ((ct-bt)*1000000 > AH401B_TIMEOUT)
	{
		Tango::Except::throw_exception(
			(const char *) "Failed to retreive data",
			(const char *) "Timeout",
			(const char *) "AH401B::read_CH3Data()",Tango::ERR);
	}

	*attr_CH3Data_read = (double) data;
	if (attr_Background_write)
		*attr_CH3Data_read -= background_level[2];
	if (attr_Calibration_write)
		*attr_CH3Data_read *= cal_matrix[2][attr_Range_write];

	attr.set_value(attr_CH3Data_read);		
	
}

//+----------------------------------------------------------------------------
//
// method : 		AH401B::read_CH4Data
// 
// description : 	Extract real attribute values for CH4Data acquisition result.
//
//-----------------------------------------------------------------------------
void AH401B::read_CH4Data(Tango::Attribute &attr)
{
	//DEBUG_STREAM << "AH401B::read_CH4Data(Tango::Attribute &attr) entering... "<< endl;

	long ret,data;
	unsigned long bunchnumber;
	struct timeval buf_time,cur_time;
	
	gettimeofday(&cur_time,NULL);
	
	ret = read_last(&data, &buf_time, &bunchnumber, buf[3]);
	double bt = buf_time.tv_sec + ((double)buf_time.tv_usec)/1000000;
	double ct = cur_time.tv_sec + ((double)cur_time.tv_usec)/1000000;	
	//DEBUG_STREAM << "AH401B::read_CH4Data() buf= "<<hex<<(unsigned int)buf[3]<<dec<<" buf_time.tv_sec="<<(unsigned int)buf_time.tv_sec <<" data="<<(double)data<<endl;
	//if (((cur_time.tv_sec-buf_time.tv_sec)*1000000+(cur_time.tv_usec-buf_time.tv_usec)) > AH401B_TIMEOUT)
	if ((ct-bt)*1000000 > AH401B_TIMEOUT)
	{
		Tango::Except::throw_exception(
			(const char *) "Failed to retreive data",
			(const char *) "Timeout",
			(const char *) "AH401B::read_CH4Data()",Tango::ERR);
	}

	*attr_CH4Data_read = (double) data;
	if (attr_Background_write)
		*attr_CH4Data_read -= background_level[3];
	if (attr_Calibration_write)
		*attr_CH4Data_read *= cal_matrix[3][attr_Range_write];			
	
	attr.set_value(attr_CH4Data_read);			
	
}

//+----------------------------------------------------------------------------
//
// method : 		AH401B::read_CH1
// 
// description : 	Extract real attribute values for CH1 acquisition result.
//
//-----------------------------------------------------------------------------
void AH401B::read_CH1(Tango::Attribute &attr)
{
	//DEBUG_STREAM << "AH401B::read_CH1(Tango::Attribute &attr) entering... "<< endl;
	
	long ret,data;
	unsigned long bunchnumber;
	struct timeval buf_time,cur_time;
	
	gettimeofday(&cur_time,NULL);
	
	ret = read_last(&data, &buf_time, &bunchnumber, buf[0]);
	double bt = buf_time.tv_sec + ((double)buf_time.tv_usec)/1000000;
	double ct = cur_time.tv_sec + ((double)cur_time.tv_usec)/1000000;
	//DEBUG_STREAM << "AH401B::read_CH1() buf= "<<hex<<(unsigned int)buf[0]<<dec<<" buf_time.tv_sec="<<(unsigned int)buf_time.tv_sec <<" data="<<(double)data<<endl;
	//if (((cur_time.tv_sec-buf_time.tv_sec)*1000000+(cur_time.tv_usec-buf_time.tv_usec)) > AH401B_TIMEOUT)
	if ((ct-bt)*1000000 > AH401B_TIMEOUT)
	{
		//DEBUG_STREAM << "AH401B::read_CH1() TIMEOUT buf time="<<fixed<<bt<<" cur time="<<ct<<" diff="<<ct-bt<<" data="<<(double)data<<endl;
		Tango::Except::throw_exception(
			(const char *) "Failed to retreive data",
			(const char *) "Timeout",
			(const char *) "AH401B::read_CH1()",Tango::ERR);
	}
	//DEBUG_STREAM << "AH401B::read_CH1Data() OK buf time="<<fixed<<bt<<" cur time="<<ct<<" diff="<<ct-bt<<" data="<<(double)data<<endl;
	*attr_CH1_read = (double) data;// - 4096;
	if (attr_Background_write)
		*attr_CH1_read -= background_level[0];
	*attr_CH1_read *= ((double)val_ranges[attr_Range_write]/pow(2.0,20));
	
	attr.set_value(attr_CH1_read);
}

//+----------------------------------------------------------------------------
//
// method : 		AH401B::read_CH2
// 
// description : 	Extract real attribute values for CH2 acquisition result.
//
//-----------------------------------------------------------------------------
void AH401B::read_CH2(Tango::Attribute &attr)
{
	//DEBUG_STREAM << "AH401B::read_CH2Data(Tango::Attribute &attr) entering... "<< endl;
	
	long ret,data;
	unsigned long bunchnumber;
	struct timeval buf_time,cur_time;
	
	gettimeofday(&cur_time,NULL);
	
	ret = read_last(&data, &buf_time, &bunchnumber, buf[1]);
	double bt = buf_time.tv_sec + ((double)buf_time.tv_usec)/1000000;
	double ct = cur_time.tv_sec + ((double)cur_time.tv_usec)/1000000;
	//DEBUG_STREAM << "AH401B::read_CH2Data() buf= "<<hex<<(unsigned int)buf[0]<<dec<<" buf_time.tv_sec="<<(unsigned int)buf_time.tv_sec <<" data="<<(double)data<<endl;
	//if (((cur_time.tv_sec-buf_time.tv_sec)*1000000+(cur_time.tv_usec-buf_time.tv_usec)) > AH401B_TIMEOUT)
	if ((ct-bt)*1000000 > AH401B_TIMEOUT)
	{
		//DEBUG_STREAM << "AH401B::read_CH2Data() TIMEOUT buf time="<<fixed<<bt<<" cur time="<<ct<<" diff="<<ct-bt<<" data="<<(double)data<<endl;
		Tango::Except::throw_exception(
			(const char *) "Failed to retreive data",
			(const char *) "Timeout",
			(const char *) "AH401B::read_CH2()",Tango::ERR);
	}
	//DEBUG_STREAM << "AH401B::read_CH2Data() OK buf time="<<fixed<<bt<<" cur time="<<ct<<" diff="<<ct-bt<<" data="<<(double)data<<endl;
	*attr_CH2_read = (double) data;// - 4096;
	if (attr_Background_write)
		*attr_CH2_read -= background_level[1];
	*attr_CH2_read *= ((double)val_ranges[attr_Range_write]/pow(2.0,20));
	
	attr.set_value(attr_CH2_read);
}

//+----------------------------------------------------------------------------
//
// method : 		AH401B::read_CH3
// 
// description : 	Extract real attribute values for CH3 acquisition result.
//
//-----------------------------------------------------------------------------
void AH401B::read_CH3(Tango::Attribute &attr)
{
	//DEBUG_STREAM << "AH401B::read_CH3Data(Tango::Attribute &attr) entering... "<< endl;
	
	long ret,data;
	unsigned long bunchnumber;
	struct timeval buf_time,cur_time;
	
	gettimeofday(&cur_time,NULL);
	
	ret = read_last(&data, &buf_time, &bunchnumber, buf[2]);
	double bt = buf_time.tv_sec + ((double)buf_time.tv_usec)/1000000;
	double ct = cur_time.tv_sec + ((double)cur_time.tv_usec)/1000000;
	//DEBUG_STREAM << "AH401B::read_CH3Data() buf= "<<hex<<(unsigned int)buf[0]<<dec<<" buf_time.tv_sec="<<(unsigned int)buf_time.tv_sec <<" data="<<(double)data<<endl;
	//if (((cur_time.tv_sec-buf_time.tv_sec)*1000000+(cur_time.tv_usec-buf_time.tv_usec)) > AH401B_TIMEOUT)
	if ((ct-bt)*1000000 > AH401B_TIMEOUT)
	{
		//DEBUG_STREAM << "AH401B::read_CH3Data() TIMEOUT buf time="<<fixed<<bt<<" cur time="<<ct<<" diff="<<ct-bt<<" data="<<(double)data<<endl;
		Tango::Except::throw_exception(
			(const char *) "Failed to retreive data",
			(const char *) "Timeout",
			(const char *) "AH401B::read_CH3()",Tango::ERR);
	}
	//DEBUG_STREAM << "AH401B::read_CH3Data() OK buf time="<<fixed<<bt<<" cur time="<<ct<<" diff="<<ct-bt<<" data="<<(double)data<<endl;
	*attr_CH3_read = (double) data;// - 4096;
	if (attr_Background_write)
		*attr_CH3_read -= background_level[2];
	*attr_CH3_read *= ((double)val_ranges[attr_Range_write]/pow(2.0,20));
	
	attr.set_value(attr_CH3_read);
}

//+----------------------------------------------------------------------------
//
// method : 		AH401B::read_CH4
// 
// description : 	Extract real attribute values for CH4 acquisition result.
//
//-----------------------------------------------------------------------------
void AH401B::read_CH4(Tango::Attribute &attr)
{
	//DEBUG_STREAM << "AH401B::read_CH4Data(Tango::Attribute &attr) entering... "<< endl;
	
	long ret,data;
	unsigned long bunchnumber;
	struct timeval buf_time,cur_time;
	
	gettimeofday(&cur_time,NULL);
	
	ret = read_last(&data, &buf_time, &bunchnumber, buf[3]);
	double bt = buf_time.tv_sec + ((double)buf_time.tv_usec)/1000000;
	double ct = cur_time.tv_sec + ((double)cur_time.tv_usec)/1000000;
	//DEBUG_STREAM << "AH401B::read_CH4Data() buf= "<<hex<<(unsigned int)buf[0]<<dec<<" buf_time.tv_sec="<<(unsigned int)buf_time.tv_sec <<" data="<<(double)data<<endl;
	//if (((cur_time.tv_sec-buf_time.tv_sec)*1000000+(cur_time.tv_usec-buf_time.tv_usec)) > AH401B_TIMEOUT)
	if ((ct-bt)*1000000 > AH401B_TIMEOUT)
	{
		//DEBUG_STREAM << "AH401B::read_CH4Data() TIMEOUT buf time="<<fixed<<bt<<" cur time="<<ct<<" diff="<<ct-bt<<" data="<<(double)data<<endl;
		Tango::Except::throw_exception(
			(const char *) "Failed to retreive data",
			(const char *) "Timeout",
			(const char *) "AH401B::read_CH4()",Tango::ERR);
	}
	//DEBUG_STREAM << "AH401B::read_CH4Data() OK buf time="<<fixed<<bt<<" cur time="<<ct<<" diff="<<ct-bt<<" data="<<(double)data<<endl;
	*attr_CH4_read = (double) data;// - 4096;
	if (attr_Background_write)
		*attr_CH4_read -= background_level[3];
	*attr_CH4_read *= ((double)val_ranges[attr_Range_write]/pow(2.0,20));
	
	attr.set_value(attr_CH4_read);
}

//+----------------------------------------------------------------------------
//
// method : 		AH401B::read_CH1Offset
// 
// description : 	Extract real attribute values for CH1Offset acquisition result.
//
//-----------------------------------------------------------------------------
void AH401B::read_CH1Offset(Tango::Attribute &attr)
{
	//DEBUG_STREAM << "AH401B::read_CH1Data(Tango::Attribute &attr) entering... "<< endl;
	*attr_CH1Offset_read = background_level[0];
	if (attr_Calibration_write)
		*attr_CH1Offset_read *= cal_matrix[0][attr_Range_write];
	attr.set_value(attr_CH1Offset_read);	
	
}

//+----------------------------------------------------------------------------
//
// method : 		AH401B::write_CH1Offset
// 
// description : 	Write CH1Offset attribute values to hardware.
//
//-----------------------------------------------------------------------------
void AH401B::write_CH1Offset(Tango::WAttribute &attr)
{
	DEBUG_STREAM << "AH401B::write_CH1Offset(Tango::WAttribute &attr) entering... "<< endl;

	attr.get_write_value(attr_CH1Offset_write);
	background_level[0] = attr_CH1Offset_write;
	if (attr_Calibration_write)
		background_level[0] /= cal_matrix[0][attr_Range_write];

	char conf_str[100];
	memset(conf_str,0,100);
	sprintf(conf_str,"%ld:%ld:%ld:%ld",(long)background_level[0],(long)background_level[1],(long)background_level[2],(long)background_level[3]);

	//salvataggio proprieta' usando host_rw e port_rw per connettersi al database
	Tango::Database *db = new Tango::Database(host_rw,port_rw);
	try {
		string configuration(conf_str);
		Tango::DbDatum config("BackgroundLevel");
		Tango::DbData db_data_del, db_data_put;
		db_data_del.push_back(config);
		config << configuration;
		db_data_put.push_back(config);
		db->put_device_property(get_name(), db_data_put);
	}
	catch (CORBA::Exception &e) {
		Tango::Except::throw_exception (
			(const char *)"Failed to update configuration",
			(const char *)"Can't write into database",
			(const char *)"AH401B::write_CH1()");
	}
	delete db;

}

//+----------------------------------------------------------------------------
//
// method : 		AH401B::read_CH2Offset
// 
// description : 	Extract real attribute values for CH2Offset acquisition result.
//
//-----------------------------------------------------------------------------
void AH401B::read_CH2Offset(Tango::Attribute &attr)
{
	//DEBUG_STREAM << "AH401B::read_CH2Data(Tango::Attribute &attr) entering... "<< endl;
	*attr_CH2Offset_read = background_level[1];
	if (attr_Calibration_write)
		*attr_CH2Offset_read *= cal_matrix[1][attr_Range_write];
	attr.set_value(attr_CH2Offset_read);	
	
}

//+----------------------------------------------------------------------------
//
// method : 		AH401B::write_CH2Offset
// 
// description : 	Write CH2Offset attribute values to hardware.
//
//-----------------------------------------------------------------------------
void AH401B::write_CH2Offset(Tango::WAttribute &attr)
{
	DEBUG_STREAM << "AH401B::write_CH2Offset(Tango::WAttribute &attr) entering... "<< endl;

	attr.get_write_value(attr_CH2Offset_write);
	background_level[1] = attr_CH2Offset_write;
	if (attr_Calibration_write)
		background_level[1] /= cal_matrix[1][attr_Range_write];
		
	char conf_str[100];
	memset(conf_str,0,100);
	sprintf(conf_str,"%ld:%ld:%ld:%ld",(long)background_level[0],(long)background_level[1],(long)background_level[2],(long)background_level[3]);

	//salvataggio proprieta' usando host_rw e port_rw per connettersi al database
	Tango::Database *db = new Tango::Database(host_rw,port_rw);
	try {
		string configuration(conf_str);
		Tango::DbDatum config("BackgroundLevel");
		Tango::DbData db_data_del, db_data_put;
		db_data_del.push_back(config);
		config << configuration;
		db_data_put.push_back(config);
		db->put_device_property(get_name(), db_data_put);
	}
	catch (CORBA::Exception &e) {
		Tango::Except::throw_exception (
			(const char *)"Failed to update configuration",
			(const char *)"Can't write into database",
			(const char *)"AH401B::write_CH2()");
	}
	delete db;
}

//+----------------------------------------------------------------------------
//
// method : 		AH401B::read_CH3Offset
// 
// description : 	Extract real attribute values for CH3Offset acquisition result.
//
//-----------------------------------------------------------------------------
void AH401B::read_CH3Offset(Tango::Attribute &attr)
{
	//DEBUG_STREAM << "AH401B::read_CH3Data(Tango::Attribute &attr) entering... "<< endl;
	*attr_CH3Offset_read = background_level[2];
	if (attr_Calibration_write)
		*attr_CH3Offset_read *= cal_matrix[2][attr_Range_write];
	attr.set_value(attr_CH3Offset_read);	
	
}

//+----------------------------------------------------------------------------
//
// method : 		AH401B::write_CH3Offset
// 
// description : 	Write CH3Offset attribute values to hardware.
//
//-----------------------------------------------------------------------------
void AH401B::write_CH3Offset(Tango::WAttribute &attr)
{
	DEBUG_STREAM << "AH401B::write_CH3Offset(Tango::WAttribute &attr) entering... "<< endl;

	attr.get_write_value(attr_CH3Offset_write);
	background_level[2] = attr_CH3Offset_write;
	if (attr_Calibration_write)
		background_level[2] /= cal_matrix[2][attr_Range_write];

	char conf_str[100];
	memset(conf_str,0,100);
	sprintf(conf_str,"%ld:%ld:%ld:%ld",(long)background_level[0],(long)background_level[1],(long)background_level[2],(long)background_level[3]);

	//salvataggio proprieta' usando host_rw e port_rw per connettersi al database
	Tango::Database *db = new Tango::Database(host_rw,port_rw);
	try {
		string configuration(conf_str);
		Tango::DbDatum config("BackgroundLevel");
		Tango::DbData db_data_del, db_data_put;
		db_data_del.push_back(config);
		config << configuration;
		db_data_put.push_back(config);
		db->put_device_property(get_name(), db_data_put);
	}
	catch (CORBA::Exception &e) {
		Tango::Except::throw_exception (
			(const char *)"Failed to update configuration",
			(const char *)"Can't write into database",
			(const char *)"AH401B::write_CH3()");
	}
	delete db;
}

//+----------------------------------------------------------------------------
//
// method : 		AH401B::read_CH4Offset
// 
// description : 	Extract real attribute values for CH4Offset acquisition result.
//
//-----------------------------------------------------------------------------
void AH401B::read_CH4Offset(Tango::Attribute &attr)
{
	//DEBUG_STREAM << "AH401B::read_CH4Data(Tango::Attribute &attr) entering... "<< endl;
	*attr_CH4Offset_read = background_level[3];
	if (attr_Calibration_write)
		*attr_CH4Offset_read *= cal_matrix[3][attr_Range_write];
	attr.set_value(attr_CH4Offset_read);	
	
}

//+----------------------------------------------------------------------------
//
// method : 		AH401B::write_CH4Offset
// 
// description : 	Write CH4Offset attribute values to hardware.
//
//-----------------------------------------------------------------------------
void AH401B::write_CH4Offset(Tango::WAttribute &attr)
{
	DEBUG_STREAM << "AH401B::write_CH4Offset(Tango::WAttribute &attr) entering... "<< endl;

	attr.get_write_value(attr_CH4Offset_write);
	background_level[3] = attr_CH4Offset_write;
	if (attr_Calibration_write)
		background_level[3] /= cal_matrix[3][attr_Range_write];

	char conf_str[100];
	memset(conf_str,0,100);
	sprintf(conf_str,"%ld:%ld:%ld:%ld",(long)background_level[0],(long)background_level[1],(long)background_level[2],(long)background_level[3]);

	//salvataggio proprieta' usando host_rw e port_rw per connettersi al database
	Tango::Database *db = new Tango::Database(host_rw,port_rw);
	try {
		string configuration(conf_str);
		Tango::DbDatum config("BackgroundLevel");
		Tango::DbData db_data_del, db_data_put;
		db_data_del.push_back(config);
		config << configuration;
		db_data_put.push_back(config);
		db->put_device_property(get_name(), db_data_put);
	}
	catch (CORBA::Exception &e) {
		Tango::Except::throw_exception (
			(const char *)"Failed to update configuration",
			(const char *)"Can't write into database",
			(const char *)"AH401B::write_CH4()");
	}
	delete db;
}


//+----------------------------------------------------------------------------
//
// method : 		AH401B::read_PAStatus
// 
// description : 	Extract real attribute values for PAStatus acquisition result.
//
//-----------------------------------------------------------------------------
void AH401B::read_PAStatus(Tango::Attribute &attr)
{
	//DEBUG_STREAM << "AH401B::read_PAStatus(Tango::Attribute &attr) entering... "<< endl;
    attr.set_value(attr_PAStatus_read, 8);
}

void AH401B::update_StatArray(bool flag_itm_upd, bool flag_rng_upd, bool flag_naq_upd, bool flag_alarm1_upd, bool flag_alarm2_upd)
{
    // #define BDR 2
/*    sendCmd("BDR ?");
    getResponse();
	if(strstr(response,"BDR") == NULL)
		getResponse();
	if(strstr(response,"BDR") != NULL)
	{
		strcpy(tmp, response+4*sizeof(char));
		statusArray[BDR] = atoi(tmp);
	}


    // #define BIN 2
    sendCmd("BIN ?");
    getResponse();
	if(strstr(response,"BIN") == NULL)
		getResponse();
	if(strstr(response,"BIN") != NULL)
	{
		if (!strcmp(response, "BIN OFF"))
		statusArray[BIN] = 0;
		else
		statusArray[BIN] = 1;
	}

// #define HLF 3
    sendCmd("HLF ?");
    getResponse();
	if(strstr(response,"HLF") == NULL)
		getResponse();
	if(strstr(response,"HLF") != NULL)
	{
		if (!strcmp(response, "HLF OFF"))
		statusArray[HLF] = 0;
		else
		statusArray[HLF] = 1;
	}*/

// #define ITM 4
	mutex->lock();
	if(get_state() == Tango::OFF || flag_itm_upd)
	{
		int count_resp = 0;
		do
		{
			if(count_resp == 1)
				usleep(50000);
			sendCmd("ITM ?");
			getResponse();
			if(strstr(response,"ITM") == NULL && strstr(response,"NAK") == NULL)
				getResponse();
			count_resp++;
		} while(strstr(response,"NAK") != NULL && count_resp < 2);
		if(strstr(response,"ITM") != NULL)
		{
			strcpy(tmp, response+4*sizeof(char));
			//statusArray[ITM] = atoi(tmp)/10;
			*attr_IntegrationTime_read = atof(tmp)/10.0;
		}
		else
			//statusArray[ITM] = -1;
			*attr_IntegrationTime_read = -1.0;
	}
	
	// #define RNG 5
	if(get_state() == Tango::OFF || flag_rng_upd)
	{
		int count_resp = 0;
		do
		{
			if(count_resp == 1)
				usleep(50000);
			sendCmd("RNG ?");
			getResponse();
			if(strstr(response,"RNG") == NULL && strstr(response,"NAK") == NULL)
				getResponse();
			count_resp++;
		} while(strstr(response,"NAK") != NULL && count_resp < 2);
		if(strstr(response,"RNG") != NULL)
		{
			strcpy(tmp, response+4*sizeof(char));
			statusArray[RNG] = atoi(tmp);
		}
		else
			statusArray[RNG] = -1;
	}

	if(get_state() == Tango::OFF || flag_naq_upd)
	{
		int count_resp = 0;
		do
		{
			if(count_resp == 1)
				usleep(50000);
			sendCmd("NAQ ?");
			getResponse();
			if(strstr(response,"NAQ") == NULL && strstr(response,"NAK") == NULL)
				getResponse();
			count_resp++;
		} while(strstr(response,"NAK") != NULL && count_resp < 2);
		if(strstr(response,"NAQ") != NULL)
		{
			strcpy(tmp, response+4*sizeof(char));
			statusArray[NAQ] = atoi(tmp);
		}
		else
			statusArray[NAQ] = -1;
	}

	if((get_state() == Tango::OFF || flag_alarm1_upd) && xpiMode)
	{
		int count_resp = 0;
		do
		{
			if(count_resp == 1)
				usleep(50000);
			sendCmd("ALM1:?");
			getResponse();
			if(strstr(response,"ALM1") == NULL && strstr(response,"NAK") == NULL)
				getResponse();
			count_resp++;
		} while(strstr(response,"NAK") != NULL && count_resp < 2);
		DEBUG_STREAM << "ALM1: read response="<<response;
		if(strstr(response,"ALM1") != NULL)
		{
			strcpy(tmp, response+5*sizeof(char));
			*attr_Alarm1_read = atoi(tmp);
		}
		else
			*attr_Alarm1_read = -1;
		DEBUG_STREAM << "ALM1: converted response="<<*attr_Alarm1_read;
	}

	if((get_state() == Tango::OFF || flag_alarm2_upd) && xpiMode)
	{
		int count_resp = 0;
		do
		{
			if(count_resp == 1)
				usleep(50000);
			sendCmd("ALM2:?");
			getResponse();
			if(strstr(response,"ALM2") == NULL && strstr(response,"NAK") == NULL)
				getResponse();
			count_resp++;
		} while(strstr(response,"NAK") != NULL && count_resp < 2);
		DEBUG_STREAM << "ALM2: read response="<<response; 
		if(strstr(response,"ALM2") != NULL)
		{
			strcpy(tmp, response+5*sizeof(char));
			*attr_Alarm2_read = atoi(tmp);
		}
		else
			*attr_Alarm2_read = -1;
		DEBUG_STREAM << "ALM2: converted response="<<*attr_Alarm2_read; 
	}

// #define TRG 6
	if(triggered)
	{
		int count_resp = 0;
		do
		{
			if(count_resp == 1)
				usleep(50000);
			sendCmd("TRG ?");
			getResponse();
			if(strstr(response,"TRG") == NULL && strstr(response,"NAK") == NULL)
				getResponse();
			count_resp++;
		} while(strstr(response,"NAK") != NULL && count_resp < 2);
		if(strstr(response,"TRG") != NULL)
		{
			//if (!strcmp(response, "TRG OFF"))
			if (strstr(response, "OFF") != NULL)
			{
				statusArray[TRG] = 0;
				if(get_state() == Tango::ON)	//not if FAULT
				{
					Tango::DevState st = Tango::STANDBY;
					set_state(st);
#if _PUSH_EVENTS
					try {
						push_change_event("state", &st);
					} catch(Tango::DevFailed &e)
					{
						INFO_STREAM << "AH401B::update_StatArray() error pushing state="<<e.errors[0].desc<<endl;
					}
#endif
					set_status("Not acquiring");
				}
			}
			else
			{
				statusArray[TRG] = 1;
				if(get_state() == Tango::STANDBY)	//not if FAULT
				{
					Tango::DevState st = Tango::ON;
					set_state(st);
#if _PUSH_EVENTS
					try {
						push_change_event("state", &st);
					} catch(Tango::DevFailed &e)
					{
						INFO_STREAM << "AH401B::update_StatArray() error pushing state="<<e.errors[0].desc<<endl;
					}
#endif
					char tmp_stat[256];
					sprintf(tmp_stat, "Acquiring, rng=%s", (statusArray[RNG] != -1) ? ranges[statusArray[RNG]] : "unknown");
					set_status(tmp_stat);
				}
				else
				{
					string status = get_status();
					if(status.find("rng=unknown") != string::npos)
					{
						char tmp_stat[256];
						sprintf(tmp_stat, "Acquiring, rng=%s", (statusArray[RNG] != -1) ? ranges[statusArray[RNG]] : "unknown");
						set_status(tmp_stat);
					}
				}
			}
		}
	}
	else	//not triggered
	{
		int count_resp = 0;
		do
		{
			if(count_resp == 1)
				usleep(50000);
			sendCmd("ACQ ?");
			getResponse();
			if(strstr(response,"ACQ") == NULL && strstr(response,"NAK") == NULL)
				getResponse();
			count_resp++;
		} while(strstr(response,"NAK") != NULL && count_resp < 2);
		if(strstr(response,"ACQ") != NULL)
		{
			//if (!strcmp(response, "ACQ OFF"))
			if (strstr(response, "OFF") != NULL)
			{
				statusArray[ACQ] = 0;
				if(get_state() == Tango::ON)	//not if FAULT
				{
					Tango::DevState st = Tango::STANDBY;
					set_state(st);
#if _PUSH_EVENTS
					try {
						push_change_event("state", &st);
					} catch(Tango::DevFailed &e)
					{
						INFO_STREAM << "AH401B::update_StatArray() error pushing state="<<e.errors[0].desc<<endl;
					}
#endif
					set_status("Not acquiring (not triggered)");
				}
			}
			else
			{
				statusArray[ACQ] = 1;
				if(get_state() == Tango::STANDBY)	//not if FAULT
				{
					Tango::DevState st = Tango::ON;
					set_state(st);
#if _PUSH_EVENTS
					try {
						push_change_event("state", &st);
					} catch(Tango::DevFailed &e)
					{
						INFO_STREAM << "AH401B::update_StatArray() error pushing state="<<e.errors[0].desc<<endl;
					}
#endif
					char tmp_stat[256];
					double free_freq = 1000.0 / attr_IntegrationTime_write;
					free_freq = halfMode ? (free_freq/2) : free_freq;
					sprintf(tmp_stat, "Acquiring (not triggered) at %3.1f Hz, rng=%s", free_freq, (statusArray[RNG] != -1) ? ranges[statusArray[RNG]] : "unknown");
					set_status(tmp_stat);
				}
				else
				{
					string status = get_status();
					if(status.find("rng=unknown") != string::npos)
					{
						char tmp_stat[256];
						sprintf(tmp_stat, "Acquiring, rng=%s", (statusArray[RNG] != -1) ? ranges[statusArray[RNG]] : "unknown");
						set_status(tmp_stat);
					}
				}
			}
		}
	}
	mutex->unlock();

//   cout<<"UPDATE - samples_index: "<<this->samples_index<<endl;
//   cout<<"update_StatArray; flag: "<<this->abort_acquire<<endl;
	gettimeofday(&last_state_time,NULL);
}

void AH401B::update_HV(bool flag_hvs_upd, bool flag_hvs_set_upd)
{
	mutex->lock();
	if(get_state() == Tango::STANDBY || flag_hvs_set_upd )
	{
		int count_resp = 0;
		do
		{
			if(count_resp == 1)
				usleep(50000);
			sendCmd("HVS:?");
			getResponse();
			if(strstr(response,"HVS") == NULL && strstr(response,"NAK") == NULL)
				getResponse();
			count_resp++;
		} while(strstr(response,"NAK") != NULL && count_resp < 2);
		if(strstr(response,"HVS") != NULL)
		{
			strcpy(tmp, response+4*sizeof(char));
			*attr_HighVoltageSet_read = atof(tmp);
		}
		/*else
			*attr_HighVoltageSet_read = -1;*/
	}
	if(get_state() != Tango::FAULT)
	{
		int count_resp = 0;
		do
		{
			if(count_resp == 1)
				usleep(50000);
			sendCmd("HVR:?");
			getResponse();
			if(strstr(response,"HVR") == NULL && strstr(response,"NAK") == NULL)
				getResponse();
			count_resp++;
		} while(strstr(response,"NAK") != NULL && count_resp < 2);
		if(strstr(response,"HVR") != NULL)
		{
			strcpy(tmp, response+4*sizeof(char));
			*attr_HighVoltage_read = atof(tmp);
		}
		/*else
			*attr_HighVoltageSet_read = -1;*/
	}
	mutex->unlock();
	gettimeofday(&last_hv_time,NULL);
}


void AH401B::sendCmd (Tango::DevString cmd)
{
	string command (cmd);
	if(xpiMode)
		replace( command.begin(), command.end(), ' ', ':');
	DEBUG_STREAM << "AH401B::sendCmd: sending command: '" << command << "'";
	command += CR;
	//command += LF;

	//send_udp_socket((char *)command.c_str());
	picoammeter_send_command(picoammeterIndex, command.c_str(), picoammeter);
}


int AH401B::getResponse()
{
	int nByte;
	int count = 0;
	int ret;
	usleep(100000);
	memset(response, 0, BUF_LEN);
	ret = picoammeter_read_response(picoammeterIndex,response,picoammeter);
	while(ret != 0 && count < 6)
	{
		usleep(100000);
		ret = picoammeter_read_response(picoammeterIndex,response,picoammeter);
		count++;
	}
	if (ret == 0)
	{
		DEBUG_STREAM << "AH401B::getResponse: response is: '"<<response<<"'";
		nByte = strlen(response);
		comm_error_mask |= COMM_OK;	//add to COMM_ERROR if present
		state_tout = STATE_TOUT;
		return nByte;
	}
	else 
	{
		DEBUG_STREAM << "AH401B::getResponse: response queue empty!";
		if(get_state() != Tango::FAULT)
		{
			Tango::DevState st = Tango::FAULT;
			set_state(st);
#if _PUSH_EVENTS
			try {
				push_change_event("state", &st);
			} catch(Tango::DevFailed &e)
			{
				INFO_STREAM << "AH401B::getResponse() error pushing state="<<e.errors[0].desc<<endl;
			}
#endif
			set_status("Trigger missing or hw fault");
		}
		comm_error_mask = COMM_ERROR;	//remove COMM_OK
		return 0;
	}

}

//+------------------------------------------------------------------
/**
 *	method:	AH401B::send_command
 *
 *	description:	method to execute "SendCommand"
 *	Send any command you want and see what happens.
 *
 * @param	argin	Command string
 * @return	Reply
 *
 */
//+------------------------------------------------------------------
Tango::DevString AH401B::send_command(Tango::DevString argin)
{
	//	POGO has generated a method core with argout allocation.
	//	If you would like to use a static reference without copying,
	//	See "TANGO Device Server Programmer's Manual"
	//		(chapter : Writing a TANGO DS / Exchanging data)
	//------------------------------------------------------------
	//Tango::DevString ret = new char[BUF_LEN+1];
	Tango::DevString ret = CORBA::string_alloc(BUF_LEN+1);
	if(strlen(argin) > 15)
	{
		INFO_STREAM << __FUNCTION__ << " command too long! " << strlen(argin);
        	Tango::Except::throw_exception(
			(const char *) "Bad Command",
			(const char *) "Command too long",
			(const char *) "AH401B::send_command()", Tango::ERR);
	}
	string cmd(argin);
	
	try {
		mutex->lock();
		int count_resp = 0;
		do
		{
			if(count_resp == 1)
				usleep(50000);
			sendCmd(argin);
			if(cmd.find("GET") == string::npos)
				getResponse();
			else
				memset(response, 0, BUF_LEN);
			count_resp++;
		} while(strstr(response,"NAK") != NULL && count_resp < 2);
		mutex->unlock();
	}
	catch(Tango::DevFailed e) {
		mutex->unlock();
		TangoSys_MemStream out_stream;
		out_stream << "AH401B::send_command error" << ends;
		WARN_STREAM << "AH401B::send_command error" << endl;
		Tango::Except::re_throw_exception(e,
			(const char *) "Communication error",
			out_stream.str(),
			(const char *) "AH401B::send_command()", Tango::ERR);
    }
    strncpy(ret, response, BUF_LEN);
//     cout<<"-----------> RET: "<<ret<<endl;
    return ret;
}


//+------------------------------------------------------------------
/**
 *	method:	AH401B::get_samples
 *
 *	description:	method to execute "GetSamples"
 *	Returns last acquired samples (length depends on NSamples property).
 *
 * @param	argin	Fake. Use 1 and get all 4 channels
 * @return	Array of samples
 *
 */
//+------------------------------------------------------------------
Tango::DevVarULongArray *AH401B::get_samples(Tango::DevUShort argin)
{
	//	POGO has generated a method core with argout allocation.
	//	If you would like to use a static reference without copying,
	//	See "TANGO Device Server Programmer's Manual"
	//		(chapter : Writing a TANGO DS / Exchanging data)
	//------------------------------------------------------------
	DEBUG_STREAM << "AH401B::get_samples(): entering... !" << endl;

    Tango::DevVarULongArray	*argout  = new Tango::DevVarULongArray();
    argout->length((nSamples*4));
    Tango::DevLong i;

    for (i=1; i<=nSamples; i++)
    {
      if ((samples_index-i) >=0) {
//         cout<<">argout["<<nSamples-i<<"] = "<<"buffer1["<<samples_index-i<<"]"<<endl;
        (*argout)[nSamples-i] = (*samplesBuffer1)[samples_index-i];
      }
      else {
//         cout<<"argout["<<nSamples-i<<"] = "<<"buffer1["<<nSamples+samples_index-i<<"]"<<endl;
        (*argout)[nSamples-i] = (*samplesBuffer1)[nSamples+samples_index-i];
      }
    }

    for (i=1; i<=nSamples; i++)
    {
      if ((samples_index-i) >=0) {
//         cout<<">argout["<<2*nSamples-i<<"] = "<<"buffer2["<<samples_index-i<<"]"<<endl;
        (*argout)[2*nSamples-i] = (*samplesBuffer2)[samples_index-i];
      }
      else {
//         cout<<"argout["<<2*nSamples-i<<"] = "<<"buffer2["<<nSamples+samples_index-i<<"]"<<endl;
        (*argout)[2*nSamples-i] = (*samplesBuffer2)[nSamples+samples_index-i];
      }
    }

    for (i=1; i<=nSamples; i++)
    {
      if ((samples_index-i) >=0) {
        (*argout)[3*nSamples-i] = (*samplesBuffer3)[samples_index-i];
//         cout<<">argout["<<nSamples-i<<"] = "<<"buffer["<<samples_index-i<<"]"<<endl;
      }
      else {
        (*argout)[3*nSamples-i] = (*samplesBuffer3)[nSamples+samples_index-i];
//         cout<<"argout["<<nSamples-i<<"] = "<<"buffer["<<nSamples+samples_index-i<<"]"<<endl;
      }
    }

    for (i=1; i<=nSamples; i++)
    {
      if ((samples_index-i) >=0) {
        (*argout)[4*nSamples-i] = (*samplesBuffer4)[samples_index-i];
//         cout<<">argout["<<nSamples-i<<"] = "<<"buffer["<<samples_index-i<<"]"<<endl;
      }
      else {
        (*argout)[4*nSamples-i] = (*samplesBuffer4)[nSamples+samples_index-i];
//         cout<<"argout["<<nSamples-i<<"] = "<<"buffer["<<nSamples+samples_index-i<<"]"<<endl;
      }
    }
    return argout;
}

//+------------------------------------------------------------------
/**
 *	method:	AH401B::get_snapshot
 *
 *	description:	method to execute "GetSnapshot"
 *	Read a single snapshot (4 channels)
 *
 * @return	The array with the 4 current measurements
 *
 */
//+------------------------------------------------------------------
Tango::DevVarLongArray *AH401B::get_snapshot()
{
	//	POGO has generated a method core with argout allocation.
	//	If you would like to use a static reference without copying,
	//	See "TANGO Device Server Programmer's Manual"
	//		(chapter : Writing a TANGO DS / Exchanging data)
	//------------------------------------------------------------
	Tango::DevVarLongArray	*argout  = new Tango::DevVarLongArray();
	argout->length(4);
	DEBUG_STREAM << "AH401B::get_snapshot(): entering... !" << endl;

	try {
		mutex->lock();
		sendCmd("?");
		//getResponse();	//response is in CHxData attributes
		mutex->unlock();
	}
	catch(Tango::DevFailed e) {
		TangoSys_MemStream out_stream;
		out_stream << "AH401B::get_snapshot error" << ends;
		WARN_STREAM << "AH401B::get_snapshot error" << endl;
		Tango::Except::re_throw_exception(e,
			(const char *) "Communication error",
			out_stream.str(),
			(const char *) "AH401B::get_snapshot()", Tango::ERR);
	}
	usleep(500000);
	long ret,data[4];
	unsigned long bunchnumber;
	struct timeval buf_time[4],cur_time;
	
	gettimeofday(&cur_time,NULL);
	
	ret = read_last(&data[0], &buf_time[0], &bunchnumber, buf[0]);
	ret = read_last(&data[1], &buf_time[1], &bunchnumber, buf[1]);
	ret = read_last(&data[2], &buf_time[2], &bunchnumber, buf[2]);
	ret = read_last(&data[3], &buf_time[3], &bunchnumber, buf[3]);

	if ((((cur_time.tv_sec-buf_time[0].tv_sec)*100000+(cur_time.tv_usec-buf_time[0].tv_usec)) > 1000000) ||
		(((cur_time.tv_sec-buf_time[1].tv_sec)*100000+(cur_time.tv_usec-buf_time[1].tv_usec)) > 1000000) ||
		(((cur_time.tv_sec-buf_time[2].tv_sec)*100000+(cur_time.tv_usec-buf_time[2].tv_usec)) > 1000000) ||
		(((cur_time.tv_sec-buf_time[3].tv_sec)*100000+(cur_time.tv_usec-buf_time[3].tv_usec)) > 1000000))
	{
		Tango::Except::throw_exception(
		(const char *) "Failed to retreive data",
		(const char *) "Timeout",
		(const char *) "AH401B::get_snapshot()",Tango::ERR);
	}

	for(int i=0; i<PICO_NUM_CHAN; i++)
	{
		(*argout)[i] =data[i];
		/*if (attr_Background_write)
			(*argout)[i]  -= background_level[i];
		if (attr_Calibration_write)
			(*argout)[i]  *= cal_matrix[i][attr_Range_write];*/
	}

/*	long resp[4];
	int ret_sscanf = sscanf(response,"%d %d %d %d",&resp[0],&resp[1],
							&resp[2],&resp[3]);
	if(ret_sscanf == 4)
	{
		for(int i=0; i<PICO_NUM_CHAN; i++)
			(*argout)[i] =resp[i];
	}
	else
	{
		TangoSys_MemStream out_stream;
		out_stream << "Failed to parse response, found " << ret_sscanf << " channel data instead of 4" << ends;
		Tango::Except::throw_exception(
			(const char *) "No response",
			out_stream.str(),
			(const char *) "AH401B::get_snapshot()", Tango::ERR);
	}*/
/*    tok = strtok(response," ");
//     cout<<"strtol1:"<<strtol(tok, NULL, 10)<<endl;
    (*argout)[0] = strtol(tok, NULL, 10);

    tok = strtok(NULL," ");
    (*argout)[1] = strtol(tok, NULL, 10);

    tok = strtok(NULL," ");
    (*argout)[2] = strtol(tok, NULL, 10);

    tok = strtok(NULL," ");
    (*argout)[3] = strtol(tok, NULL, 10);*/

	return argout;
}



//+------------------------------------------------------------------
/**
 *	method:	AH401B::set_integration_time
 *
 *	description:	method to execute "SetIntegrationTime"
 *	Sets the integration time (range is the same defined in the manual).
 *
 * @param	argin	Integration time (10-10000)
 *
 */
//+------------------------------------------------------------------
void AH401B::set_integration_time(Tango::DevDouble argin)
{
	DEBUG_STREAM << "AH401B::set_integration_time(): entering... !" << endl;

	char cmd[8];
	long val = (long)(argin * 10);
	if ((val<10) || (val>10000))
	{
		Tango::Except::throw_exception (
			(const char *)"PARAMETER ERROR",
			(const char *)"Parameter out of range [1,1000] ms",
			(const char *)"AH401B::set_integration_time()");
	}

	ostringstream lt;
	lt<<"ITM "<<val;
	strcpy(cmd, lt.str().c_str());
	Tango::DevString ret;
	try {
		ret = this->send_command(cmd);
	}
	catch(Tango::DevFailed e)
	{
		TangoSys_MemStream out_stream;
		out_stream << "AH401B::set_integration_time error" << ends;
		WARN_STREAM << "AH401B::set_integration_time error" << endl;
		CORBA::string_free(ret);
		Tango::Except::re_throw_exception(e,
			(const char *) "Communication error",
			out_stream.str(),
			(const char *) "AH401B::set_integration_time()", Tango::ERR);
	}
	CORBA::string_free(ret);
	Tango::MultiAttribute *attr_list = get_device_attr();
	Tango::WAttribute &attr = attr_list->get_w_attr_by_name ("IntegrationTime");
	attr_IntegrationTime_write = ((double)val)/10;
	attr.set_write_value(attr_IntegrationTime_write);
	flag_itm_upd = true;

	if(!triggered && get_state()==Tango::ON)
	{
		char tmp_stat[256];
		double free_freq = 1000.0 / attr_IntegrationTime_write;
		free_freq = halfMode ? (free_freq/2) : free_freq;
		sprintf(tmp_stat, "Acquiring (not triggered) at %3.1f Hz, rng=%s", free_freq, (statusArray[RNG] != -1) ? ranges[statusArray[RNG]] : "unknown");
		set_status(tmp_stat);
	}
	else if(triggered && get_state()==Tango::ON)
	{
		char tmp_stat[256];
		sprintf(tmp_stat, "Acquiring, rng=%s", (statusArray[RNG] != -1) ? ranges[statusArray[RNG]] : "unknown");
		set_status(tmp_stat);
	}
}

//+------------------------------------------------------------------
/**
 *	method:	AH401B::set_alarm1
 *
 */
//+------------------------------------------------------------------
void AH401B::set_alarm1(Tango::DevLong argin)
{
	DEBUG_STREAM << "AH401B::set_alarm1(): entering... !" << endl;

	char cmd[32];
	long val = argin;
	if (!xpiMode)
	{
		Tango::Except::throw_exception(
                        (const char *) "Not Allowed",
                        (const char *) "Allowed only for Xpi",
                        (const char *)__FUNCTION__,Tango::ERR);
	}

	ostringstream lt;
	lt<<"ALM1 "<<val;
	strcpy(cmd, lt.str().c_str());
	Tango::DevString ret;
	try {
		ret = this->send_command(cmd);
	}
	catch(Tango::DevFailed e)
	{
		TangoSys_MemStream out_stream;
		out_stream << "AH401B::set_alarm1 error" << ends;
		WARN_STREAM << "AH401B::set_alarm1 error" << endl;
		CORBA::string_free(ret);
		Tango::Except::re_throw_exception(e,
			(const char *) "Communication error",
			out_stream.str(),
			(const char *) "AH401B::set_alarm1()", Tango::ERR);
	}
	CORBA::string_free(ret);
	try
	{
		Tango::MultiAttribute *attr_list = get_device_attr();
		Tango::WAttribute &attr = attr_list->get_w_attr_by_name ("Alarm1");
		attr_Alarm1_write = val;
		attr.set_write_value(attr_Alarm1_write);
	}
	catch(...)
	{
		INFO_STREAM << __FUNCTION__ << " exception setting write value";
	} 
	
	flag_alarm1_upd = true;
}

//+------------------------------------------------------------------
/**
 *	method:	AH401B::set_alarm2
 *
 */
//+------------------------------------------------------------------
void AH401B::set_alarm2(Tango::DevLong argin)
{
	DEBUG_STREAM << "AH401B::set_alarm2(): entering... !" << endl;

	char cmd[32];
	long val = argin;
	if (!xpiMode)
	{
		Tango::Except::throw_exception(
                        (const char *) "Not Allowed",
                        (const char *) "Allowed only for Xpi",
                        (const char *)__FUNCTION__,Tango::ERR);
	}

	ostringstream lt;
	lt<<"ALM2 "<<val;
	strcpy(cmd, lt.str().c_str());
	Tango::DevString ret;
	try {
		ret = this->send_command(cmd);
	}
	catch(Tango::DevFailed e)
	{
		TangoSys_MemStream out_stream;
		out_stream << "AH401B::set_alarm2 error" << ends;
		WARN_STREAM << "AH401B::set_alarm2 error" << endl;
		CORBA::string_free(ret);
		Tango::Except::re_throw_exception(e,
			(const char *) "Communication error",
			out_stream.str(),
			(const char *) "AH401B::set_alarm2()", Tango::ERR);
	}
	CORBA::string_free(ret);
	try
	{
		Tango::MultiAttribute *attr_list = get_device_attr();
		Tango::WAttribute &attr = attr_list->get_w_attr_by_name ("Alarm2");
		attr_Alarm2_write = val;
		attr.set_write_value(attr_Alarm2_write);
	}
        catch(...)
        {
                INFO_STREAM << __FUNCTION__ << " exception setting write value";
        }
	flag_alarm2_upd = true;
}

//+------------------------------------------------------------------
/**
 *	method:	AH401B::set_full_scale_range
 *
 *	description:	method to execute "SetFullScaleRange"
 *	Scale range codes as in the manual.
 *
 * @param	argin	Scale range (0-7)
 *
 */
//+------------------------------------------------------------------
void AH401B::set_full_scale_range(Tango::DevUShort argin)
{
	DEBUG_STREAM << "AH401B::set_full_scale_range(): entering... !" << endl;
	char cmd[8];
	if (argin>7)
	{
	Tango::Except::throw_exception (
		(const char *)"PARAMETER ERROR",
		(const char *)"Parameter out of range",
		(const char *)"AH401B::set_full_scale_range()");
	}
	
	ostringstream lt;
	lt<<"RNG "<<argin;
	strcpy(cmd, lt.str().c_str());
	Tango::DevString ret;
	try {
		ret = this->send_command(cmd);
	}
	catch(Tango::DevFailed e) {
		TangoSys_MemStream out_stream;
		out_stream << "AH401B::set_full_scale_range error" << ends;
		WARN_STREAM << "AH401B::set_full_scale_range error" << endl;
	CORBA::string_free(ret);
		Tango::Except::re_throw_exception(e,
			(const char *) "Communication error",
			out_stream.str(),
			(const char *) "AH401B::set_full_scale_range()", Tango::ERR);
	}
	CORBA::string_free(ret);
	Tango::MultiAttribute *attr_list = get_device_attr();
	Tango::WAttribute &attr = attr_list->get_w_attr_by_name ("Range");
	attr_Range_write = argin;
	attr.set_write_value(attr_Range_write);
	flag_rng_upd = true;

	if(!triggered && get_state()==Tango::ON)
	{
		char tmp_stat[256];
		double free_freq = 1000.0 / attr_IntegrationTime_write;
		free_freq = halfMode ? (free_freq/2) : free_freq;
		sprintf(tmp_stat, "Acquiring (not triggered) at %3.1f Hz, rng=%s", free_freq, (statusArray[RNG] != -1) ? ranges[statusArray[RNG]] : "unknown");
		set_status(tmp_stat);
	}
	else if(triggered && get_state()==Tango::ON)
	{
		char tmp_stat[256];
		sprintf(tmp_stat, "Acquiring, rng=%s", (statusArray[RNG] != -1) ? ranges[statusArray[RNG]] : "unknown");
		set_status(tmp_stat);
	}
}

//+------------------------------------------------------------------
/**
 *	method:	AH401B::set_num_acq
 *
 *	description:	method to execute "NumAcq"
 *	Scale range codes as in the manual.
 *
 * @param	argin	Number of acquisition per trigger
 *
 */
//+------------------------------------------------------------------
void AH401B::set_num_acq(Tango::DevLong argin)
{
	DEBUG_STREAM << "AH401B::set_num_acq(): entering... !" << endl;
	char cmd[8];
	if (argin>7)
	{
		Tango::Except::throw_exception (
			(const char *)"PARAMETER ERROR",
			(const char *)"Parameter out of range",
			(const char *)"AH401B::set_num_acq()");
	}

	ostringstream lt;
	lt<<"NAQ "<<argin;
	strcpy(cmd, lt.str().c_str());
	Tango::DevString ret;
	try {
		ret = this->send_command(cmd);
	}
	catch(Tango::DevFailed e) {
		TangoSys_MemStream out_stream;
		out_stream << "AH401B::set_num_acq error" << ends;
		WARN_STREAM << "AH401B::set_num_acq error" << endl;
		CORBA::string_free(ret);
		Tango::Except::re_throw_exception(e,
			(const char *) "Communication error",
			out_stream.str(),
			(const char *) "AH401B::set_num_acq()", Tango::ERR);
	}
	CORBA::string_free(ret);
	Tango::MultiAttribute *attr_list = get_device_attr();
	Tango::WAttribute &attr = attr_list->get_w_attr_by_name ("NumAcq");
	attr_NumAcq_write = argin;
	attr.set_write_value(attr_NumAcq_write);
	flag_naq_upd = true;
}

//+------------------------------------------------------------------
/**
 *	method:	AH401B::set_high_volt
 *
 *	description:	method to execute "HighVoltage"
 *	Scale range codes as in the manual.
 *
 * @param	argin	High Voltage
 *
 */
//+------------------------------------------------------------------
void AH401B::set_high_volt(Tango::DevDouble argin)
{
	DEBUG_STREAM << "AH401B::set_high_volt(): entering... !" << endl;
	char cmd[8];
	/*if (argin>100)	//TODO
	{
		Tango::Except::throw_exception (
			(const char *)"PARAMETER ERROR",
			(const char *)"Parameter out of range",
			(const char *)"AH401B::set_high_volt()");
	}*/

	ostringstream lt;
	lt<<"HVS:"<<argin;
	strcpy(cmd, lt.str().c_str());

	try {
		this->send_command(cmd);
	}
	catch(Tango::DevFailed e) {
		TangoSys_MemStream out_stream;
		out_stream << "AH401B::set_high_volt error" << ends;
		WARN_STREAM << "AH401B::set_high_volt error" << endl;
		Tango::Except::re_throw_exception(e,
			(const char *) "Communication error",
			out_stream.str(),
			(const char *) "AH401B::set_high_volt()", Tango::ERR);
	}
	/*Tango::MultiAttribute *attr_list = get_device_attr();
	Tango::WAttribute &attr = attr_list->get_w_attr_by_name ("NumAcq");
	attr_NumAcq_write = argin;
	attr.set_write_value(attr_NumAcq_write);*/
	//usleep(1000000);
	gettimeofday(&last_hv_set_time, NULL);
	flag_hvs_upd = true;
	flag_hvs_set_upd = true;
}

//+------------------------------------------------------------------
/**
 *	method:	AH401B::start
 *
 *	description:	method to execute "Start"
 *	Start acquisition on trigger.
 *
 * @return	
 *
 */
//+------------------------------------------------------------------
void AH401B::start()
{
	DEBUG_STREAM << "AH401B::start(): entering... !" << endl;

	char cmd[8];
	if(triggered)
	{
		ostringstream lt;
		lt<<"TRG ON";
		strcpy(cmd, lt.str().c_str());
		Tango::DevString ret;
		try {
			ret = this->send_command(cmd);
		}
		catch(Tango::DevFailed e) {
			TangoSys_MemStream out_stream;
			out_stream << "AH401B::start error" << ends;
			WARN_STREAM << "AH401B::start error" << endl;
			CORBA::string_free(ret);
			Tango::Except::re_throw_exception(e,
				(const char *) "Communication error",
				out_stream.str(),
				(const char *) "AH401B::start()", Tango::ERR);
		}
		CORBA::string_free(ret);
		flag_trg_upd = true;
	}
	else
	{
		attr_NumAcq_write = 0;
		set_num_acq(attr_NumAcq_write);
		
		ostringstream lt;
		lt<<"ACQ ON";
		strcpy(cmd, lt.str().c_str());
		Tango::DevString ret;
		try
		{
			ret = this->send_command(cmd);
		}
		catch(Tango::DevFailed e)
		{
			TangoSys_MemStream out_stream;
			out_stream << "AH401B::start error" << ends;
			WARN_STREAM << "AH401B::start error" << endl;
			CORBA::string_free(ret);
			Tango::Except::re_throw_exception(e,
				(const char *) "Communication error",
				out_stream.str(),
				(const char *) "AH401B::start()", Tango::ERR);
		}
		CORBA::string_free(ret);
		flag_acq_upd = true;
	}
}

//+------------------------------------------------------------------
/**
 *	method:	AH401B::stop
 *
 *	description:	method to execute "Stop"
 *	Start acquisition on trigger.
 *
 * @return	
 *
 */
//+------------------------------------------------------------------
void AH401B::stop()
{
	DEBUG_STREAM << "AH401B::stop(): entering... !" << endl;

	char cmd[8];

	ostringstream lt;
	if(triggered)
		lt<<"TRG OFF";
	else
		lt<<"ACQ OFF";
	strcpy(cmd, lt.str().c_str());
	Tango::DevString ret;
	try {
		ret = this->send_command(cmd);
	}
	catch(Tango::DevFailed e) {
		TangoSys_MemStream out_stream;
		out_stream << "AH401B::stop error" << ends;
		WARN_STREAM << "AH401B::stop error" << endl;
		CORBA::string_free(ret);
		Tango::Except::re_throw_exception(e,
			(const char *) "Communication error",
			out_stream.str(),
			(const char *) "AH401B::stop()", Tango::ERR);
	}
	CORBA::string_free(ret);
	if(triggered)
		flag_trg_upd = true;
	else
		flag_acq_upd = true;
	Tango::DevState st = Tango::OFF;
	set_state(st);
#if _PUSH_EVENTS
	try {
		push_change_event("state", &st);
	} catch(Tango::DevFailed &e)
	{
		INFO_STREAM << "AH401B::stop() error pushing state="<<e.errors[0].desc<<endl;
	}
#endif
	set_status("Not acquiring");
}


//+------------------------------------------------------------------
/**
 *	method:	AH401B::get_ch1_data
 *
 *	description:	method to execute "GetCH1Data"
 *
 * @param	argin	
 * @return	
 *
 */
//+------------------------------------------------------------------
Tango::DevVarDoubleArray *AH401B::get_ch1_data(const Tango::DevVarLongArray *argin)
{
	//	POGO has generated a method core with argout allocation.
	//	If you would like to use a static reference without copying,
	//	See "TANGO Device Server Programmer's Manual"
	//		(chapter : Writing a TANGO DS / Exchanging data)
	//------------------------------------------------------------
	Tango::DevVarDoubleArray	*argout  = new Tango::DevVarDoubleArray();

	DEBUG_STREAM << "AH401B::get_ch1_data(): entering... !" << endl;

	//	Add your own code to control device here
	long buf_idx[2], num_samples;
	unsigned char unique;

	//DEBUG_STREAM << __FUNCTION__<<": not timeout!" << endl;
	try
	{
		check_get_data_param(argin, buf[0], buf_idx, &num_samples, &unique);
	}
	catch(Tango::DevFailed e)
	{
		string err = string(e.errors[0].desc);
		DEBUG_STREAM << __FUNCTION__<<" Exception in check_get_data_param="<<err;
		delete argout;
		Tango::Except::re_throw_exception(e,
			(const char *) "Failed to retreive data",
			err,
			(const char *) __FUNCTION__,Tango::ERR);
	}
	argout->length(num_samples);
	
	/* get last value */
	//read_data_calib(long idx[2], double *data_buf, struct _buffer* buffer, double x2, double x1, double x0)
	double a = 1,b = 0;	
	if (attr_Background_write) {
		if (attr_Calibration_write) {
			a = cal_matrix[0][attr_Range_write]; b = -background_level[0]*a;
		}
		else {
			b = -background_level[0];
		}
	}
	else {
		if (attr_Calibration_write) {
			a = cal_matrix[0][attr_Range_write];
		}
	}

	read_data_calib(unique, buf_idx, &(*argout)[0], buf[0], 0, a, b);

	return argout;
	
}

//+------------------------------------------------------------------
/**
 *	method:	AH401B::get_ch2_data
 *
 *	description:	method to execute "GetCH2Data"
 *
 * @param	argin	
 * @return	
 *
 */
//+------------------------------------------------------------------
Tango::DevVarDoubleArray *AH401B::get_ch2_data(const Tango::DevVarLongArray *argin)
{
	//	POGO has generated a method core with argout allocation.
	//	If you would like to use a static reference without copying,
	//	See "TANGO Device Server Programmer's Manual"
	//		(chapter : Writing a TANGO DS / Exchanging data)
	//------------------------------------------------------------
	Tango::DevVarDoubleArray	*argout  = new Tango::DevVarDoubleArray();

	DEBUG_STREAM << "AH401B::get_ch2_data(): entering... !" << endl;

	//	Add your own code to control device here
	long buf_idx[2], num_samples;
	unsigned char unique;

	try
	{
		check_get_data_param(argin, buf[1], buf_idx, &num_samples, &unique);
	}
	catch(Tango::DevFailed e)
	{
		string err = string(e.errors[0].desc);
		DEBUG_STREAM << __FUNCTION__<<" Exception in check_get_data_param="<<err;
		delete argout;
		Tango::Except::re_throw_exception(e,
			(const char *) "Failed to retreive data",
			err,
			(const char *) __FUNCTION__,Tango::ERR);
	}
	argout->length(num_samples);

	/* get last value */
	//read_data_calib(long idx[2], double *data_buf, struct _buffer* buffer, double x2, double x1, double x0)
	
	double a = 1,b = 0;	
	if (attr_Background_write) {
		if (attr_Calibration_write) {
			a = cal_matrix[1][attr_Range_write]; b = -background_level[1]*a;
		}
		else {
			b = -background_level[1];
		}
	}
	else {
		if (attr_Calibration_write) {
			a = cal_matrix[1][attr_Range_write];
		}
	}

	read_data_calib(unique, buf_idx, &(*argout)[0], buf[1], 0, a, b);

	return argout;	
	
}

//+------------------------------------------------------------------
/**
 *	method:	AH401B::get_ch3_data
 *
 *	description:	method to execute "GetCH3Data"
 *
 * @param	argin	
 * @return	
 *
 */
//+------------------------------------------------------------------
Tango::DevVarDoubleArray *AH401B::get_ch3_data(const Tango::DevVarLongArray *argin)
{
	//	POGO has generated a method core with argout allocation.
	//	If you would like to use a static reference without copying,
	//	See "TANGO Device Server Programmer's Manual"
	//		(chapter : Writing a TANGO DS / Exchanging data)
	//------------------------------------------------------------
	Tango::DevVarDoubleArray	*argout  = new Tango::DevVarDoubleArray();
	DEBUG_STREAM << "AH401B::get_ch3_data(): entering... !" << endl;

	//	Add your own code to control device here
	long buf_idx[2], num_samples;
	unsigned char unique;

	try
	{
		check_get_data_param(argin, buf[2], buf_idx, &num_samples, &unique);
	}
	catch(Tango::DevFailed e)
	{
		string err = string(e.errors[0].desc);
		DEBUG_STREAM << __FUNCTION__<<" Exception in check_get_data_param="<<err;
		delete argout;
		Tango::Except::re_throw_exception(e,
			(const char *) "Failed to retreive data",
			err,
			(const char *) __FUNCTION__,Tango::ERR);
	}
	argout->length(num_samples);

	/* get last value */
	//read_data_calib(long idx[2], double *data_buf, struct _buffer* buffer, double x2, double x1, double x0)
	double a = 1,b = 0;	
	if (attr_Background_write) {
		if (attr_Calibration_write) {
			a = cal_matrix[2][attr_Range_write]; b = -background_level[2]*a;
		}
		else {
			b = -background_level[2];
		}
	}
	else {
		if (attr_Calibration_write) {
			a = cal_matrix[2][attr_Range_write];
		}
	}

	read_data_calib(unique, buf_idx, &(*argout)[0], buf[2], 0, a, b);
	return argout;
}

//+------------------------------------------------------------------
/**
 *	method:	AH401B::get_ch4_data
 *
 *	description:	method to execute "GetCH4Data"
 *
 * @param	argin	
 * @return	
 *
 */
//+------------------------------------------------------------------
Tango::DevVarDoubleArray *AH401B::get_ch4_data(const Tango::DevVarLongArray *argin)
{
	//	POGO has generated a method core with argout allocation.
	//	If you would like to use a static reference without copying,
	//	See "TANGO Device Server Programmer's Manual"
	//		(chapter : Writing a TANGO DS / Exchanging data)
	//------------------------------------------------------------
	Tango::DevVarDoubleArray	*argout  = new Tango::DevVarDoubleArray();
	DEBUG_STREAM << "AH401B::get_ch4_data(): entering... !" << endl;

	//	Add your own code to control device here
	long buf_idx[2], num_samples;
	unsigned char unique;

	try
	{
		check_get_data_param(argin, buf[3], buf_idx, &num_samples, &unique);
	}
	catch(Tango::DevFailed e)
	{
		string err = string(e.errors[0].desc);
		DEBUG_STREAM << __FUNCTION__<<" Exception in check_get_data_param="<<err;
		delete argout;
		Tango::Except::re_throw_exception(e,
			(const char *) "Failed to retreive data",
			err,
			(const char *) __FUNCTION__,Tango::ERR);
	}
	argout->length(num_samples);

	/* get last value */
	//read_data_calib(long idx[2], double *data_buf, struct _buffer* buffer, double x2, double x1, double x0)
	double a = 1,b = 0;	
	if (attr_Background_write) {
		if (attr_Calibration_write) {
			a = cal_matrix[3][attr_Range_write]; b = -background_level[3]*a;
		}
		else {
			b = -background_level[3];
		}
	}
	else {
		if (attr_Calibration_write) {
			a = cal_matrix[3][attr_Range_write];
		}
	}
	read_data_calib(unique, buf_idx, &(*argout)[0], buf[3], 0, a, b);
	return argout;
}

//+------------------------------------------------------------------
/**
 *	method:	AH401B::get_ch1
 *
 *	description:	method to execute "GetCH1"
 *
 * @param	argin	
 * @return	
 *
 */
//+------------------------------------------------------------------
Tango::DevVarDoubleArray *AH401B::get_ch1(const Tango::DevVarLongArray *argin)
{
	//	POGO has generated a method core with argout allocation.
	//	If you would like to use a static reference without copying,
	//	See "TANGO Device Server Programmer's Manual"
	//		(chapter : Writing a TANGO DS / Exchanging data)
	//------------------------------------------------------------
	Tango::DevVarDoubleArray	*argout  = new Tango::DevVarDoubleArray();

	DEBUG_STREAM << "AH401B::get_ch1(): entering... !" << endl;

	//	Add your own code to control device here
	long buf_idx[2], num_samples;
	unsigned char unique;

	try
	{
		check_get_data_param(argin, buf[0], buf_idx, &num_samples, &unique);
	}
	catch(Tango::DevFailed e)
	{
		string err = string(e.errors[0].desc);
		DEBUG_STREAM << __FUNCTION__<<" Exception in check_get_data_param="<<err;
		delete argout;
		Tango::Except::re_throw_exception(e,
			(const char *) "Failed to retreive data",
			err,
			(const char *) __FUNCTION__,Tango::ERR);
	}
	argout->length(num_samples);
	
	/* get last value */
	//read_data_calib(long idx[2], double *data_buf, struct _buffer* buffer, double x2, double x1, double x0)
	double a = 1,b = 0;
	a = val_ranges[attr_Range_write]/pow(2.0,20);
	if (attr_Background_write)
		 b = -background_level[0]*a;
	else
		b = 0;
	read_data_calib(unique, buf_idx, &(*argout)[0], buf[0], 0, a, b);

	return argout;
}

//+------------------------------------------------------------------
/**
 *	method:	AH401B::get_ch2
 *
 *	description:	method to execute "GetCH2"
 *
 * @param	argin	
 * @return	
 *
 */
//+------------------------------------------------------------------
Tango::DevVarDoubleArray *AH401B::get_ch2(const Tango::DevVarLongArray *argin)
{
	//	POGO has generated a method core with argout allocation.
	//	If you would like to use a static reference without copying,
	//	See "TANGO Device Server Programmer's Manual"
	//		(chapter : Writing a TANGO DS / Exchanging data)
	//------------------------------------------------------------
	Tango::DevVarDoubleArray	*argout  = new Tango::DevVarDoubleArray();

	DEBUG_STREAM << "AH401B::get_ch2(): entering... !" << endl;

	//	Add your own code to control device here
	long buf_idx[2], num_samples;
	unsigned char unique;

	try
	{
		check_get_data_param(argin, buf[1], buf_idx, &num_samples, &unique);
	}
	catch(Tango::DevFailed e)
	{
		string err = string(e.errors[0].desc);
		DEBUG_STREAM << __FUNCTION__<<" Exception in check_get_data_param="<<err;
		delete argout;
		Tango::Except::re_throw_exception(e,
			(const char *) "Failed to retreive data",
			err,
			(const char *) __FUNCTION__,Tango::ERR);
	}
	argout->length(num_samples);
	
	/* get last value */
	//read_data_calib(long idx[2], double *data_buf, struct _buffer* buffer, double x2, double x1, double x0)
	double a = 1,b = 0;
	a = val_ranges[attr_Range_write]/pow(2.0,20);
	if (attr_Background_write)
		 b = -background_level[1]*a;
	else
		b = 0;
	read_data_calib(unique, buf_idx, &(*argout)[0], buf[1], 0, a, b);

	return argout;
}

//+------------------------------------------------------------------
/**
 *	method:	AH401B::get_ch3
 *
 *	description:	method to execute "GetCH3"
 *
 * @param	argin	
 * @return	
 *
 */
//+------------------------------------------------------------------
Tango::DevVarDoubleArray *AH401B::get_ch3(const Tango::DevVarLongArray *argin)
{
	//	POGO has generated a method core with argout allocation.
	//	If you would like to use a static reference without copying,
	//	See "TANGO Device Server Programmer's Manual"
	//		(chapter : Writing a TANGO DS / Exchanging data)
	//------------------------------------------------------------
	Tango::DevVarDoubleArray	*argout  = new Tango::DevVarDoubleArray();

	DEBUG_STREAM << "AH401B::get_ch3(): entering... !" << endl;

	//	Add your own code to control device here
	long buf_idx[2], num_samples;
	unsigned char unique;

	try
	{
		check_get_data_param(argin, buf[2], buf_idx, &num_samples, &unique);
	}
	catch(Tango::DevFailed e)
	{
		string err = string(e.errors[0].desc);
		DEBUG_STREAM << __FUNCTION__<<" Exception in check_get_data_param="<<err;
		delete argout;
		Tango::Except::re_throw_exception(e,
			(const char *) "Failed to retreive data",
			err,
			(const char *) __FUNCTION__,Tango::ERR);
	}
	argout->length(num_samples);
	
	/* get last value */
	//read_data_calib(long idx[2], double *data_buf, struct _buffer* buffer, double x2, double x1, double x0)
	double a = 1,b = 0;
	a = val_ranges[attr_Range_write]/pow(2.0,20);
	if (attr_Background_write)
		 b = -background_level[2]*a;
	else
		b = 0;
	read_data_calib(unique, buf_idx, &(*argout)[0], buf[2], 0, a, b);

	return argout;
}

//+------------------------------------------------------------------
/**
 *	method:	AH401B::get_ch4
 *
 *	description:	method to execute "GetCH4"
 *
 * @param	argin	
 * @return	
 *
 */
//+------------------------------------------------------------------
Tango::DevVarDoubleArray *AH401B::get_ch4(const Tango::DevVarLongArray *argin)
{
	//	POGO has generated a method core with argout allocation.
	//	If you would like to use a static reference without copying,
	//	See "TANGO Device Server Programmer's Manual"
	//		(chapter : Writing a TANGO DS / Exchanging data)
	//------------------------------------------------------------
	Tango::DevVarDoubleArray	*argout  = new Tango::DevVarDoubleArray();

	DEBUG_STREAM << "AH401B::get_ch4(): entering... !" << endl;

	//	Add your own code to control device here
	long buf_idx[2], num_samples;
	unsigned char unique;

	try
	{
		check_get_data_param(argin, buf[3], buf_idx, &num_samples, &unique);
	}
	catch(Tango::DevFailed e)
	{
		string err = string(e.errors[0].desc);
		DEBUG_STREAM << __FUNCTION__<<" Exception in check_get_data_param="<<err;
		delete argout;
		Tango::Except::re_throw_exception(e,
			(const char *) "Failed to retreive data",
			err,
			(const char *) __FUNCTION__,Tango::ERR);
	}
	argout->length(num_samples);
	
	/* get last value */
	//read_data_calib(long idx[2], double *data_buf, struct _buffer* buffer, double x2, double x1, double x0)
	double a = 1,b = 0;
	a = val_ranges[attr_Range_write]/pow(2.0,20);
	if (attr_Background_write)
		 b = -background_level[3]*a;
	else
		b = 0;
	read_data_calib(unique, buf_idx, &(*argout)[0], buf[3], 0, a, b);

	return argout;
}

//+------------------------------------------------------------------
/**
 *	method:	AH401B::init_buffers
 *
 *	description:	method to init realtime buffers
 *
 * @param	argin	
 * @return	
 *
 */
//+------------------------------------------------------------------

bool AH401B::init_buffers(void)
{

	fd_buffer[0] = open(bufferCH1Device.c_str(),O_RDWR);
	if(fd_buffer[0] < 0) {
		ERROR_STREAM << "AH401B: error opening " << bufferCH1Device << endl;
		return false;
	}
	fd_buffer[1] = open(bufferCH2Device.c_str(),O_RDWR);
	if(fd_buffer[1] < 0) {
		ERROR_STREAM << "AH401B: error opening " << bufferCH2Device << endl;
		return false;
	}
	fd_buffer[2] = open(bufferCH3Device.c_str(),O_RDWR);
	if(fd_buffer[2] < 0) {
		ERROR_STREAM << "AH401B: error opening " << bufferCH3Device << endl;
		return false;
	}
	fd_buffer[3] = open(bufferCH4Device.c_str(),O_RDWR);
	if(fd_buffer[3] < 0) {
		ERROR_STREAM << "AH401B: error opening " << bufferCH4Device << endl;
		return false;
	}			
	
	for (int i = 0; i < PICO_NUM_CHAN; i++) {	
		buf[i] = 0;
		int ret =ioctl(fd_buffer[i], BUFFER_RT_IOC_READ_VSIZE, &vsize);
		if(ret) {
			ERROR_STREAM << "AH401B:  error in ioctl! fd_buffer=" << fd_buffer[i] << " ret="<< ret << endl;
			close(fd_buffer[i]);
		}
	
		DEBUG_STREAM << "AH401B:  read from buffer vsize="<< vsize << endl;
		if(vsize <= 0) {
			ERROR_STREAM << "AH401B: error vsize="<< vsize << endl;
			close(fd_buffer[i]);
		}
	
		buf[i] = 0;
		buf[i] = (struct _buffer *) mmap(0, vsize, PROT_READ | PROT_WRITE, MAP_SHARED, fd_buffer[i], 0);
	
		if (buf[i] == NULL) {
			ERROR_STREAM << "AH401B: error in mmap!!!" << endl;
			close(fd_buffer[i]);
		}
	}
	return true;
}


//+------------------------------------------------------------------
/**
 *	method:	AH401B::delete_buffers
 *
 *	description:	method to init realtime buffers
 *
 * @param	argin	
 * @return	
 *
 */
//+------------------------------------------------------------------

void AH401B::delete_buffers(void)
{

	for (int i = 0; i < 4; i++) {
		if (buf[i] != NULL) {
			munmap((void *) buf[i], vsize);
			if (fd_buffer[i])
				close(fd_buffer[i]);
		}
	}

}

#if 0
//+------------------------------------------------------------------
/**
 *	method:	AH401B::check_get_data_param
 *
 *	description:	get data indexes
 *
 * @param	argin	
 * @return	
 *
 */
//+------------------------------------------------------------------
void AH401B::check_get_data_param(const Tango::DevVarLongArray *argin, long *buf_idx, long *data_size, unsigned char *unique)
{
	long nargin = argin->length();
	long ret_num_samples;
	long mode = (*argin)[0];
	 
	if ((*argin)[0] <= BUF_GET_TIME_MODE)
		*unique = 0;
	else
		*unique = 1;;

	/* non unique modes (0,1,2); unique modes (3,4,5) */
	if (*unique == 1)
		mode = mode - 3;

	/* get last values */
	if (nargin == 2)
	{
		long num_samples = (*argin)[1];

		if ((num_samples < 0) || (num_samples > vsize)) {
			Tango::Except::throw_exception (
				(const char *)"Failed to retreive last data",
				(const char *)"Invalid number of data requested",
				(const char *)"AH401B::check_get_data_param()");
		}

		if (mode == BUF_GET_LAST_MODE) {	
			ret_num_samples = get_last_idx(num_samples, *unique, buf_idx, buf[0]); /* buf[0]: all buffers,same clock */
		}
		else {
			Tango::Except::throw_exception (
				(const char *)"Failed to retreive last data",
				(const char *)"Wrong get mode parameter (0,1,2)",
				(const char *)"AH401B::check_get_data_param()");
		}

		if (ret_num_samples != num_samples) {
			Tango::Except::throw_exception (
				(const char *)"Failed to retreive last data",
				(const char *)"Can't retreive all data requested",
				(const char *)"AH401B::check_get_data_param()");
		}
		*data_size = ret_num_samples;
	}
	/* get bunch mode values */
	else if (nargin == 3)
	{
		long bunch_start = (*argin)[1], 
		     bunch_end = (*argin)[2];



		if (((bunch_end-bunch_start) < 0) || ((bunch_end-bunch_start) > vsize)) {
			TangoSys_MemStream out_stream;
			out_stream << "Invalid bunch numbers: bunch start=" << bunch_start << ", bunch end=" << bunch_end << ends;
			INFO_STREAM << __FUNCTION__ << " " << out_stream.str();
			Tango::Except::throw_exception (
				(const char *)"Failed to retreive bunch data",
				out_stream.str(),
				(const char *)"AH401B::check_get_data_param()");
		}

		if (mode == BUF_GET_BUNCH_MODE) {
			ret_num_samples = get_bunch_idx(bunch_start, bunch_end, *unique, buf_idx, buf[0]);
		}
		else {
			TangoSys_MemStream out_stream;
			out_stream << "Invalid mode=" << mode << ends;
			Tango::Except::throw_exception (
				(const char *)"Failed to retreive bunch data",
				out_stream.str(),
				(const char *)"AH401B::check_get_data_param()");
		}

		/* check data size */
		if ((((ret_num_samples-1) != (bunch_end-bunch_start)) && *unique) ||
		    (((ret_num_samples-1) < (bunch_end-bunch_start)) && (*unique == 0))) {

			TangoSys_MemStream out_stream;
			out_stream << "Invalid bunch numbers: bunch start=" << bunch_start << ", bunch end=" << bunch_end <<  ", ret_num_samples=" <<
			(ret_num_samples-1) << ends;
			INFO_STREAM << __FUNCTION__ << " " << out_stream.str();
			Tango::Except::throw_exception (
				(const char *)"Failed to retreive bunch data",
				out_stream.str(),
				(const char *)"AH401B::check_get_data_param()");
		}
		
		*data_size = ret_num_samples;
		
	}
	/* get time mode values */
	else if (nargin == 5)
	{
		struct timeval time_start, time_end;

		time_start.tv_sec = (*argin)[1]; time_start.tv_usec = (*argin)[2];
		time_end.tv_sec = (*argin)[3]; time_end.tv_usec = (*argin)[4];

		if (((time_end.tv_sec-time_start.tv_sec)*1000000+(time_end.tv_usec-time_start.tv_usec)) < 0) {
			Tango::Except::throw_exception (
				(const char *)"Failed to retreive timed data",
				(const char *)"Invalid time slice requested",
				(const char *)"AH401B::check_get_data_param()");
		}

		if (mode == BUF_GET_TIME_MODE)
			ret_num_samples = get_time_idx(time_start, time_end, *unique, buf_idx, buf[0]);
		else {
			Tango::Except::throw_exception (
				(const char *)"Failed to retreive timed data",
				(const char *)"Wrong get mode parameter (0,1,2)",
				(const char *)"AH401B::check_get_data_param()");
		}
		*data_size = ret_num_samples;
	}
	/* invalid number of input parameters */
	else  {
		Tango::Except::throw_exception (
			(const char *)"Failed to get data",
			(const char *)"Invalid number of input parameters",
			(const char *)"AH401B::check_get_data_param()");
	}	

}
#endif

//+------------------------------------------------------------------
/**
 *	method:	AH401B::acquire_background
 *
 *	description:	method to execute "AcquireBackground"
 *	Acquire background level
 *
 *
 */
//+------------------------------------------------------------------
void AH401B::acquire_background(Tango::DevLong argin)
{
	DEBUG_STREAM << "AH401B::acquire_background(): entering... !" << endl;

	//	Add your own code to control device here
	long data[4];
	struct timeval cur_time;
	
	gettimeofday(&cur_time,NULL);

	for (int i = 0; i < PICO_NUM_CHAN; i++) {
		//ret = read_last(&data[i], &buf_time, &bunchnumber, buf[i]);
		long buf_idx[2], num_samples;
		unsigned char unique;
		double std, mean;
		Tango::DevVarLongArray	*argin2  = new Tango::DevVarLongArray();
		argin2->length(2);
		(*argin2)[0] = 0;
		(*argin2)[1] = argin;
			
		check_get_data_param(argin2, buf[i] , buf_idx, &num_samples, &unique);
		delete argin2;
		read_data_filter(unique, buf_idx, buf[i ], 0, 1, 0, &std, &mean, NULL, NULL, BUF_FILT_STD, NULL);
		
		data[i] = (long)mean;
	}
	/*if (((cur_time.tv_sec-buf_time.tv_sec)*100000+(cur_time.tv_usec-buf_time.tv_usec)) > AH401B_TIMEOUT) {
		Tango::Except::throw_exception(
		(const char *) "Failed to retreive data",
		(const char *) "Timeout",
		(const char *) "AH401B::acquire_background()",Tango::ERR);
	}*/

	char conf_str[100];
	memset(conf_str,0,100);
	sprintf(conf_str,"%ld:%ld:%ld:%ld",data[0],data[1],data[2],data[3]);

	//salvataggio proprieta' usando host_rw e port_rw per connettersi al database
	Tango::Database *db = new Tango::Database(host_rw,port_rw);
	try {
		string configuration(conf_str);
		Tango::DbDatum config("BackgroundLevel");
		Tango::DbData db_data_del, db_data_put;
		db_data_del.push_back(config);
		config << configuration;
		db_data_put.push_back(config);
		db->put_device_property(get_name(), db_data_put);
	}
	catch (CORBA::Exception &e) {
		Tango::Except::throw_exception (
			(const char *)"Failed to update configuration",
			(const char *)"Can't write into database",
			(const char *)"AH401B::acquire_background()");
	}
	delete db;

	for (int i = 0; i < PICO_NUM_CHAN; i++) {
		background_level[i] = (double)data[i];
	}

}

//+------------------------------------------------------------------
/**
 *	method:	AH401B::reset
 *
 *	description:	method to execute "Reset"
 *	Reset
 *
 *
 */
//+------------------------------------------------------------------
void AH401B::reset()
{
	DEBUG_STREAM << "AH401B::reset(): entering... !" << endl;
	if(!xpiMode)
		Tango::Except::throw_exception(
			(const char *) "Not allowed",
			(const char *) "Allowed only on Xpi",
			(const char *) "AH401B::reset()", Tango::ERR);
	
	//reset module A
	string resp;
	char cmd[16];
	strcpy(cmd, "RST:A");
	
	try {
		resp = this->send_command(cmd);
	}
	catch(Tango::DevFailed e)
	{
		TangoSys_MemStream out_stream;
		out_stream << "AH401B::set_integration_time error" << ends;
		WARN_STREAM << "AH401B::set_integration_time error" << endl;
		Tango::Except::re_throw_exception(e,
			(const char *) "Communication error",
			out_stream.str(),
			(const char *) "AH401B::reset()", Tango::ERR);
	}
	if(resp.find("NAK") != string::npos)
	{
		try {
			resp = this->send_command(cmd);
		}
		catch(Tango::DevFailed e)
		{
			TangoSys_MemStream out_stream;
			out_stream << "AH401B::set_integration_time error" << ends;
			WARN_STREAM << "AH401B::set_integration_time error" << endl;
			Tango::Except::re_throw_exception(e,
				(const char *) "Communication error",
				out_stream.str(),
				(const char *) "AH401B::reset()", Tango::ERR);
		}
	}
	
	//reset module A
	strcpy(cmd, "RST:B");
	
	try {
		resp = this->send_command(cmd);
	}
	catch(Tango::DevFailed e)
	{
		TangoSys_MemStream out_stream;
		out_stream << "AH401B::set_integration_time error" << ends;
		WARN_STREAM << "AH401B::set_integration_time error" << endl;
		Tango::Except::re_throw_exception(e,
			(const char *) "Communication error",
			out_stream.str(),
			(const char *) "AH401B::reset()", Tango::ERR);
	}
	if(resp.find("NAK") != string::npos)
	{
		try {
			resp = this->send_command(cmd);
		}
		catch(Tango::DevFailed e)
		{
			TangoSys_MemStream out_stream;
			out_stream << "AH401B::set_integration_time error" << ends;
			WARN_STREAM << "AH401B::set_integration_time error" << endl;
			Tango::Except::re_throw_exception(e,
				(const char *) "Communication error",
				out_stream.str(),
				(const char *) "AH401B::reset()", Tango::ERR);
		}
	}
	
	init_ah401b();
}

//+------------------------------------------------------------------
/**
 *	method:	AH401B::string_explode
 * Explode a string 
*/
//+------------------------------------------------------------------
void AH401B::string_explode(string str, string separator, vector<string>* results)
{
	unsigned int found;
	
	found = str.find_first_of(separator);
	while(found != string::npos) {
		if(found > 0) {
			results->push_back(str.substr(0,found));
		}
		str = str.substr(found+1);
		found = str.find_first_of(separator);
	}
	if(str.length() > 0) {
		results->push_back(str);
	}

}


//+------------------------------------------------------------------
/**
 *	method:	AH401B::string_explode
 * Explode a string 
*/
//+------------------------------------------------------------------
void AH401B::string_explode(string str, string separator, vector<unsigned int>* results, bool b_hex)
{
	unsigned int found;
	
	found = str.find_first_of(separator);
	while(found != string::npos) {
		if(found > 0) {
			unsigned int x;
			bool err = false;
			if(!b_hex)
			{
				std::istringstream i(str.substr(0,found));
				if (!(i >> x))
				{
					err = true;
				}
			}
			else
			{
				std::stringstream i; 
				i << hex << (str.substr(0,found));
				if (!(i >> x))
				{
					err = true;
				}
			}

			if (err)
			{
				DEBUG_STREAM << "AH401B::string_explode error converting '" << str << "'";
			}
			else
			{
				results->push_back(x);
				DEBUG_STREAM << "AH401B::string_explode pushing " << x << "=0x" << hex << x << dec;
			}
		}
		str = str.substr(found+1);
		found = str.find_first_of(separator);
	}
	if(str.length() > 0) {
		unsigned int x;
		bool err = false;
		if(!b_hex)
		{
			std::istringstream i(str);
			if (!(i >> x))
			{
				err = true;
			}
		}
		else
		{
			std::stringstream i; 
			i << hex << (str);
			if (!(i >> x))
			{
				err = true;
			}
		}
		if (err)
		{
			DEBUG_STREAM << "AH401B::string_explode error converting '" << str << "'";
		}
		else
		{
			results->push_back(x);
			DEBUG_STREAM << "AH401B::string_explode pushing " << x << "=0x" << hex << x << dec;
		}
	}

}

//+------------------------------------------------------------------
/**
 *	method:	AH401B::get_configuration
 * Get the pico devices
*/
//+------------------------------------------------------------------
bool AH401B::get_configuration(void)
{
	/* channel 1 */
	vector<string> results1; 
	string_explode(cH1Cal, ":", &results1);
	if (results1.size() != AH401B_NUM_RANGE) {
		cout << "Wrong configuration on channel 1" << endl;
		return false;
	}	
	for (uint k = 0; k < results1.size(); k++) {
		cal_matrix[0][k] = atof(results1[k].c_str());
	}

	/* channel 2 */
	vector<string> results2; 
	string_explode(cH2Cal, ":", &results2);
	if (results2.size() != AH401B_NUM_RANGE) {
		cout << "Wrong configuration on channel 2" << endl;
		return false;
	}	
	for (uint k = 0; k < results2.size(); k++) {
		cal_matrix[1][k] = atof(results2[k].c_str());
	}

	/* channel 3 */
	vector<string> results3; 
	string_explode(cH3Cal, ":", &results3);
	if (results3.size() != AH401B_NUM_RANGE) {
		cout << "Wrong configuration on channel 3" << endl;
		return false;
	}	
	for (uint k = 0; k < results3.size(); k++) {
		cal_matrix[2][k] = atof(results3[k].c_str());
	}

	/* channel 4 */
	vector<string> results4; 
	string_explode(cH4Cal, ":", &results4);
	if (results4.size() != AH401B_NUM_RANGE) {
		cout << "Wrong configuration on channel 4" << endl;
		return false;
	}	
	for (uint k = 0; k < results4.size(); k++) {
		cal_matrix[3][k] = atof(results4[k].c_str());
	}

	/* background */
	vector<string> results5; 
	string_explode(backgroundLevel, ":", &results5);
	if (results5.size() != PICO_NUM_CHAN) {
		cout << "Wrong configuration on background" << endl;
		return false;
	}	
	for (uint k = 0; k < results5.size(); k++) {
		background_level[k] = atof(results5[k].c_str());
	}

	return true;

}


bool AH401B::init_ah401b(void) 
{
	try {
		mutex->lock();
		sendCmd("BIN OFF");
		getResponse();

		sendCmd("ACQ OFF");
		getResponse();

		sendCmd("TRG OFF");
		getResponse();
		mutex->unlock();
	
		set_full_scale_range((short)attr_Range_write);
	
		set_integration_time(attr_IntegrationTime_write);

		set_num_acq(attr_NumAcq_write);
		
		if(xpiMode)
		{
			set_alarm1(attr_Alarm1_write);
	
			set_alarm2(attr_Alarm2_write);
		}
	
		/*sendCmd("ACQ ON");
		getResponse();*/

		mutex->lock();
		if(halfMode)
			sendCmd("HLF ON");
		else
			sendCmd("HLF OFF");
		getResponse();

		if(triggered)
			sendCmd("TRG ON");
		else
			sendCmd("ACQ ON");
		getResponse();
		mutex->unlock();

		Tango::DevState st = Tango::ON;
		set_state(st);
#if _PUSH_EVENTS
		try {
			push_change_event("state", &st);
		} catch(Tango::DevFailed &e)
		{
			INFO_STREAM << "AH401B::update_StatArray() error pushing state="<<e.errors[0].desc<<endl;
		}
#endif
		char tmp_stat[256];
		if(triggered)
		{
			sprintf(tmp_stat, "Acquiring, rng=%s", (statusArray[RNG] != -1) ? ranges[statusArray[RNG]] : "unknown");
		}
		else
		{
			double free_freq = 1000.0 / attr_IntegrationTime_write;
			free_freq = halfMode ? (free_freq/2) : free_freq;
			sprintf(tmp_stat,"Acquiring (not triggered) at %3.1f Hz, rng=%s", free_freq, (statusArray[RNG] != -1) ? ranges[statusArray[RNG]] : "unknown");
		}
		set_status(tmp_stat);
		
		return true;
	
	}
	catch(Tango::DevFailed &e) {
		printf("Initialization error\n");
			return false;			
	}
}






}	//	namespace
